#######################################################################
# makefile for compiling and linking the entire simulation
# 1) compiles the application: *.c->*.s
# 2) augments the assembly: *.s->*.S
# 3) compiles the augmented assemblies: *.s->*.o
# 4) joins all the app. object files into single archive: *.o-> <app>.a
# 5) calls the kernel makefile, which produces kernel.a
# 6) calls the simulator's makefile, which produces <sim>.a
# 7) links alltogether: kernel.a + <app>.a + <sim>.a = <app>
#######################################################################

# dirs and libs definitions
include 	$(LIMESDIR)/globals.make

# the augmentation tool
export AUG 	= $(AUGDIR)/aug	$(AUGFLAGS)

# who controles the linking process:
LDCTRL		= $(GCPP) 

# for compiling the application
APPCC		= $(GCC) $(APPCCFLAGS)

##############################################################
# the name and parameters of the instrumented standard library
AUGLEVEL=1
ifeq ($(findstring -0, $(AUGFLAGS)), -0)
AUGLEVEL=0
endif
ifeq ($(findstring -1, $(AUGFLAGS)), -1)
AUGLEVEL=1
endif
ifeq ($(findstring -2, $(AUGFLAGS)), -2)
AUGLEVEL=2
endif
export AUGLEVEL
export SYS_STDLIB=$(LIBDIR)/sys_stdlib.$(AUGLEVEL).a

#######################################################################
# This makefile is included by the application makefile.
# So the application will be the first to be compiled. Note that the
# current directory is the one where the application is.

APPSS	= $(APPCS:.c=.s)
APPSSS	= $(APPCS:.c=.S)
APPOS	= $(APPCS:.c=.o)

$(TARGET) : $(APPLIB) $(KERNELLIB) $(INTERFACELIB) $(SYNCLIB) $(SYS_STDLIB) $(SIMLIB) makefile
	$(LDCTRL) $(KERNELLIB) $(SIMLIB) $(APPLIB) $(INTERFACELIB) $(SYNCLIB) $(SYS_STDLIB) $(APPLDFLAGS) -o $(TARGET)

$(KERNELLIB): FORCE
	make -C $(KERNELDIR)        -f kernel.make	
$(SYNCLIB): FORCE
	make -C $(SYNCDIR)          -f $(SYNC).make	
$(SIMLIB): FORCE
	make -C $(SIMDIR)           -f $(SIM).make
$(SYS_STDLIB): FORCE
	make -C $(SYS_STDLIBDIR)    -f sys_stdlib.make
FORCE:	# make imagines this target to be updated whenever it is run

$(APPLIB) : $(APPOS)
	$(AR) ru $(APPLIB) $(APPOS)

$(APPOS): %.o : %.S
	$(AS) $< -o $@

$(APPSSS): %.S : %.s
	$(AUG) $< $@
	
$(APPSS): %.s : %.c $(INCLUDEDIR)/macros.h $(SYS_STDLIBDIR)/macros.h 
	$(APPCC) -I$(INCLUDEDIR) -include $(INCLUDEDIR)/macros.h        \
			 -I$(SYNCDIR) -include $(SYNCDIR)/sync.h                \
			 -I$(SYS_STDLIBDIR) -include $(SYS_STDLIBDIR)/macros.h  \
			 -S $< -o $@

###################################################
# cleaning: the application, the simulator, both...

clean:	simclean appclean                           # make sim+app clean
simclean:                                           # make sim clean
	make -C $(SIMDIR)    -f $(SIM).make	clean		
appclean:                                           # make app clean
	rm -rf $(APPOS) $(APPSS) $(APPSSS) $(APPLIB) $(TARGET)
appsclean:                                          # delete asm files
	rm -rf $(APPSS)
appSclean:                                          # delete augmented assembly files
	rm -rf $(APPSSS)
kerclean:                                           # clean up kernel directory
	make -C $(KERNELDIR) -f kernel.make	clean
syncclean:                                          # clean up sync directory
	make -C $(SYNCDIR) -f $(SYNC).make	clean
libclean:                                           # delete all "cache" .a files	
	rm -rf $(LIBDIR)/*.a
stdlibclean:
	make -C $(SYS_STDLIBDIR) -f sys_stdlib.make clean
totalclean:	clean kerclean syncclean libclean stdlibclean	# clean up everything


###########################################################
# making dependencies: for the simulator and for the kernel

dep:
	make -C $(KERNELDIR) -f kernel.make         dep
	make -C $(SYNCDIR)   -f $(SYNC).make        dep
	make -C $(SIMDIR)    -f $(SIM).make         dep
	make -C $(SYS_STDLIBDIR) -f sys_stdlib.make dep
