/*
 * Title:	Ex1 application: modification of Jacobi algoroithm
 * Author:	Aleksandar Milenkovic
 * Date:	February, 1998.
 * Note:	
 * 
 *------------------------------------------------------------
 *    Command Line Options
 *     -pP: P = number of processors; must be a power of 2.
 *     -mM: M = number of rows (even integer).
 *     -nN: N = number of columns (even integers).
 *     -iI: I = number of iteration.
 *     -h : Print out command line options.
 *------------------------------------------------------------ 
 * */
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PREF(X)  	asm(INSTRUCTION_PREF " %0" : "=m" (X));
#define PREFEX(X)	asm(INSTRUCTION_PREFEX " %0" : "=m" (X));
#define LPREF(X)	asm(INSTRUCTION_LPREF " %0" : "=m" (X));
#define WRBACK(X)	asm(INSTRUCTION_WRBACK " %0" : "=m" (X));


#define DEFAULT_P 1  


#define MAXPROCS 16
#define DEFAULT_M 64
#define DEFAULT_N 64
#define DEFAULT_I 8

int numProcs=DEFAULT_P;   
int M=DEFAULT_M;
int N=DEFAULT_N;
int numIteration=DEFAULT_I;

/* truly shared vars */
LOCKDEC(indexLock);
int Index; /* Used by processors to determine which input to take */
BARDEC(sBarrier1);
BARDEC(sBarrier2); 
double **grid;

/* private array */
double **scratch;
 
void SlaveStart();
int log_2(int);

void main(int argc, char*argv[]) {
    
    int i, j, m1;
    unsigned time;
    int c;
    extern char *optarg;
    
    while((c=getopt(argc, argv, "p:m:n:i:h")) != -1)
       {
         switch(c) {
           case 'p': numProcs=atoi(optarg);
                     if(numProcs<1) {
                       printf("P must be >=1\n");
                       exit(-1);
                     };
                     if(log_2(numProcs) == -1) {
                       printf("P must be a power of 2\n");
                       exit(-1);
                     };          	
                     break;       
           
           case 'm': M=atoi(optarg);
           	     m1 = M/2;
           	     if(2*m1 != M) {
           	        printf("M must be even\n");
           	        exit(-1);
           	     }
           	     break;
           
           case 'n': N=atoi(optarg);
                     m1 = N/2;
                     if(2*m1 != N) {
                       printf("N must be even\n");
                       exit(-1);
                     }
                     break;
                     
           case 'i': numIteration=atoi(optarg);
                     break;
                     	       
           case 'h': printf("Usage mytestp <options>\n\n");
                     printf("options:\n");
                     printf("    -pP: P = number of processors; must be a power of 2.\n");
 		     printf("    -mM: M = number of rows (even integer).\n");
                     printf("    -nN: N = number of columns (even integers).\n");
                     printf("    -iI: I = number of iterations\n");
                     printf("    -h : Print out command line options.\n");
                     printf("\nDefault: myex1 -p%d -m%d -n%d -i%d\n", DEFAULT_P, DEFAULT_M, DEFAULT_N, DEFAULT_I);   
                     exit(0);
                     break;
        }
    }

    /* Define shared matrix grid */
    grid = (double **) G_MALLOC(M*sizeof(double *));
    scratch = (double **) malloc(M*sizeof(double *));
    for(i=0; i<N; i++)
       grid[i] = (double *) G_MALLOC(N*sizeof(double));
    for(i=0; i<N; i++)
       scratch[i] = (double *) malloc(N*sizeof(double));

    /* Initialize grid */
    /*
    for(i=0; i<M; i++)
       for(j=0; j<N; j++)
          grid[i][j] = (long) (rand()/RAND_MAX);
    */
    /* Initialize the Index */
    Index = 0;

    /* Initialize the barriers and the lock */
    LOCKINIT(indexLock) 
    BARINIT(sBarrier1)
    BARINIT(sBarrier2)
        
    CLOCK(time);

    /* Create the slave processes. */
    for (i = 0; i < numProcs-1; i++)
        CREATE(SlaveStart)
        
    /* Make the master do slave work so we don't waste a processor */
    SlaveStart();        

    /* wait for all the slaves to complete */
    WAIT_FOR_END(numProcs-1)

    CLOCK(time);
    printf("Elapsed time (in cycles): %d\n", time); 
    /*
    for (i=0; i<M; i++)
      for (j=0; j<N; j++)
         printf("ADDR(grid[%d][%d])=%x\tADDR(scratch[%d][%d])=%x\n",i,j,&grid[i][j],i,j,&scratch[i][j]);
    */
       
    }


/*====================== SlaveStart =====================================*/
/* This is the routine that each processor will be executing in parallel */
void SlaveStart() {
    int myIndex, i, j, k, begin, end;
    int l, WinSize; 
    
    /*double scratch[M/numProcs][N];  scratch matrix */

    LOCK(indexLock);        /* enter the critical section */
    
    myIndex = Index;        /* read your ID */
    ++Index;                /* increment it, so the next will operate on ID+1 */

    UNLOCK(indexLock);      /* leave the critical section */
      
    /* Initialize begin and end */
    WinSize=(M/numProcs);
    begin = WinSize*myIndex;
    end = WinSize*(myIndex+1); 
      
    /* the main body of a thread */
    for(i=0; i<numIteration; ++i) {
        for(j=begin; j<end; j++) {            
            for(k=0; k<N; k++) {               
                      scratch[j][k] = 0.6*grid[j][k] + 
                      (0.4)*(grid[(j-WinSize+M)%M][k]+grid[(j+WinSize)%M][k]+
                              grid[j][(k-WinSize+N)%N]+grid[j][(k+WinSize)%N]);
                      /*
                      if (myIndex==0)
                      printf("\n<I%d P%d,%d>: s[%d][%d]  g[%d][%d]  g[%d][%d]  g[%d][%d]  g[%d][%d]", 
                             i, myIndex, l, j, k, (j-l+M)%M, k, (j+l)%M, k, j, (k-l+N)%N, j, (k+l)%N);
                      */
            } /* end of k */
        } /* end of j */
        
        BARRIER(sBarrier1, numProcs);
         
        for(j=begin; j<end; j++) {
            for(k=0; k<N; k++) {
               grid[j][k]=scratch[j][k];                          
            }
       }
       BARRIER(sBarrier2, numProcs);
       
    }

 }

/*=============== log_2 function ===============================*/
int log_2(int number)

{
  int cumulative = 1;
  int out = 0;
  int done = 0;

  while ((cumulative < number) && (!done) && (out < 50)) {
    if (cumulative == number) {
      done = 1;
    } else {
      cumulative = cumulative * 2;
      out ++;
    }
  }

  if (cumulative == number) {
    return(out);
  } else {
    return(-1);
  }
}
