/*
 * Title:	Jacobi like application
 * Author:	Aleksandar Milenkovic
 * Date:	November, 1997.
 * Note:	
 * 
 *------------------------------------------------------------
 *    Command Line Options
 *     -pP: P = number of processors; must be a power of 2.
 *     -mM: M = number of rows (even integer).
 *     -nN: N = number of columns (even integers).
 *     -h : Print out command line options.
 *------------------------------------------------------------ 
 * */
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PREF(X)  	asm(INSTRUCTION_PREF " %0" : "=m" (X));
#define PREFEX(X)	asm(INSTRUCTION_PREFEX " %0" : "=m" (X));
#define LPREF(X)	asm(INSTRUCTION_LPREF " %0" : "=m" (X));
#define WRBACK(X)	asm(INSTRUCTION_WRBACK " %0" : "=m" (X));


#define DEFAULT_P 1  


#define MAXPROCS 16
#define DEFAULT_M 64
#define DEFAULT_N 64

int numProcs=DEFAULT_P;   
int M=DEFAULT_M;
int N=DEFAULT_N;
int numIteration=4;

/* truly shared vars */
LOCKDEC(indexLock);
int Index; /* Used by processors to determine which input to take */
BARDEC(sBarrier1);
BARDEC(sBarrier2); 
long **grid;

/* private array */
long **scratch;
 
void SlaveStart();
int log_2(int);

void main(int argc, char*argv[]) {
    
    int i, j, m1;
    unsigned time;
    int c;
    extern char *optarg;
    
    while((c=getopt(argc, argv, "p:m:n:h")) != -1)
       {
         switch(c) {
           case 'p': numProcs=atoi(optarg);
                     if(numProcs<1) {
                       printf("P must be >=1\n");
                       exit(-1);
                     };
                     if(log_2(numProcs) == -1) {
                       printf("P must be a power of 2\n");
                       exit(-1);
                     };          	
                     break;       
           
           case 'm': M=atoi(optarg);
           	     m1 = M/2;
           	     if(2*m1 != M) {
           	        printf("M must be even\n");
           	        exit(-1);
           	     }
           	     break;
           
           case 'n': N=atoi(optarg);
                     m1 = N/2;
                     if(2*m1 != N) {
                       printf("N must be even\n");
                       exit(-1);
                     }
                     break;
           	       
           case 'h': printf("Usage jacobi1.out <options>\n\n");
                     printf("options:\n");
                     printf("    -pP: P = number of processors; must be a power of 2.\n");
 		     printf("    -mM: M = number of rows (even integer).\n");
                     printf("    -nN: N = number of columns (even integers).\n");
                     printf("    -h : Print out command line options.\n");
                     printf("\nDefault: jacobi1.out -p%d -m%d -n%d\n", DEFAULT_P, DEFAULT_M, DEFAULT_N);   
                     exit(0);
                     break;
        }
    }

    /* Define shared matrix grid */
    grid = (long **) G_MALLOC(M*sizeof(float *));
    scratch = (long **) malloc(M*sizeof(float *));
    for(i=0; i<N; i++)
       grid[i] = (long *) G_MALLOC(N*sizeof(float));
    for(i=0; i<N; i++)
       scratch[i] = (long *) malloc(N*sizeof(float));

    /* Initialize grid */
    /*
    for(i=0; i<M; i++)
       for(j=0; j<N; j++)
          grid[i][j] = (long) (rand()/RAND_MAX);
    */
    /* Initialize the Index */
    Index = 0;

    /* Initialize the barriers and the lock */
    LOCKINIT(indexLock) 
    BARINIT(sBarrier1)
    BARINIT(sBarrier2)
        
    CLOCK(time);

    /* Create the slave processes. */
    for (i = 0; i < numProcs-1; i++)
        CREATE(SlaveStart)
        
    /* Make the master do slave work so we don't waste a processor */
    SlaveStart();        

    /* wait for all the slaves to complete */
    WAIT_FOR_END(numProcs-1)

    CLOCK(time);
    printf("Elapsed time (in cycles): %d\n", time); 
    /*
    for (i=0; i<M; i++)
      for (j=0; j<N; j++)
         printf("ADDR(grid[%d][%d])=%x\tADDR(scratch[%d][%d])=%x\n",i,j,&grid[i][j],i,j,&scratch[i][j]);
    */
       
    }


/*====================== SlaveStart =====================================*/
/* This is the routine that each processor will be executing in parallel */
void SlaveStart() {
    int myIndex, i, j, k, begin, end;
    int l; 
    
    /*float scratch[M/numProcs][N];  scratch matrix */

    LOCK(indexLock);        /* enter the critical section */
    
    myIndex = Index;        /* read your ID */
    ++Index;                /* increment it, so the next will operate on ID+1 */

    UNLOCK(indexLock);      /* leave the critical section */
      
    /* Initialize begin and end */
    begin = (M/numProcs)*myIndex;
    end = (M/numProcs)*(myIndex+1); 
 
    /* the main body of a thread */
    for(i=0; i<numIteration; ++i) {
        l=0;	
        for(j=begin+1; j<end-1; j++) {
            
            for(k=0; k<N; k++) {               
                 scratch[j][k]=(grid[(j-1+M)%M][k]+grid[(j+1)%M][k]+
                              grid[j][(k-1+N)%N]+grid[j][(k+1)%N])/4;
                 /*              
                 if ((j==begin) && (k%8==0) && ((k+8)<N)) 
                    PREF(grid[(j-1+M)%M][(k+8)%N]);
                 if ((j==(end-1)) && (k%8==0)&& ((k+8)<N))
                    PREF(grid[(j+1)%M][(k+8)%N]);  
                */
            }
                    
            if(l<N) {
               PREF(grid[(begin-1+M)%M][l]);
               PREF(grid[end%M][l]);
            }
           l+=8;
           
        }
        for(k=0; k<N; k++) {
            scratch[begin][k]=(grid[(begin-1+M)%M][k]+grid[(begin+1)%M][k]+
                              grid[begin][(k-1+N)%N]+grid[begin][(k+1)%N])/4;
            scratch[end-1][k]=(grid[(end-2+M)%M][k]+grid[(end)%M][k]+
                              grid[end-1][(k-1+N)%N]+grid[end-1][(k+1)%N])/4;
            }

        /*
        PREF(grid[(begin-1+M)%M][0]);
        for(k=0; k<N; k++) {
            scratch[begin][k]=(grid[(begin-1+M)%M][k]+grid[(begin+1)%M][k]+
                              grid[begin][(k-1+N)%N]+grid[begin][(k+1)%N])/4;
            if ((k%8==0) && ((k+8)<N)) PREF(grid[(begin-1+M)%M][k+8]);
            }
        PREF(grid[end%M][0]);                     
        for(k=0; k<N; k++) {            
            scratch[end-1][k]=(grid[(end-2+M)%M][k]+grid[(end)%M][k]+
                              grid[end-1][(k-1+N)%N]+grid[end-1][(k+1)%N])/4;
            if ((k%8==0) && ((k+8)<N)) PREF(grid[end%M][k+8]);
        } 
        */     
        
        BARRIER(sBarrier1, numProcs);
         
        for(j=begin; j<end; j++) {
            /*if (j==end) PREFEX(grid[j][0]);*/
            for(k=0; k<N; k++) {
               /*if((j==end) && ((k+1)*8 <= N)) PREFEX(grid[j][(k+1)*8]);*/
               grid[j][k]=scratch[j][k];                          
            }
       }
       BARRIER(sBarrier2, numProcs);
       
    }

 }

/*=============== log_2 function ===============================*/
int log_2(int number)

{
  int cumulative = 1;
  int out = 0;
  int done = 0;

  while ((cumulative < number) && (!done) && (out < 50)) {
    if (cumulative == number) {
      done = 1;
    } else {
      cumulative = cumulative * 2;
      out ++;
    }
  }

  if (cumulative == number) {
    return(out);
  } else {
    return(-1);
  }
}
