/*
 * Title:	Producer Consumer kernel: PC
 * Author:	Aleksandar Milenkovic
 * Date:	March, 1998.
 * Note:	
 * 
 *------------------------------------------------------------
 *    Command Line Options
 *     -pP: P = number of processors; must be a power of 2.
 *     -iI: I = Number of Iterations
 *     -h : Print out command line options.
 *------------------------------------------------------------ 
 * */
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PREF(X)  	asm(INSTRUCTION_PREF " %0" : "=m" (X));
#define PREFEX(X)	asm(INSTRUCTION_PREFEX " %0" : "=m" (X));
#define LPREF(X)	asm(INSTRUCTION_LPREF " %0" : "=m" (X));
#define WRBACK(X)	asm(INSTRUCTION_WRBACK " %0" : "=m" (X));
#define OWL(X)          asm(INSTRUCTION_OWL " %0": "=m" (X));
#define OWH(X)          asm(INSTRUCTION_OWH " %0": "=m" (X));
#define CWL(X)          asm(INSTRUCTION_CWL " %0": "=m" (X));

#define DEFAULT_P 1  
#define MAXPROCS 16
#define DEFAULT_I 32

int numProcs=DEFAULT_P;   
int I=DEFAULT_I;

/* truly shared vars */
BARDEC(b1);
LOCKDEC(indexLock);
int Index; /* Used by processors to determine which input to take */
     

typedef struct Item 
  { double data1;
    double data2;
  } Item;

Item Var;

void ProducerStart();
void ConsumerStart();
int log_2(int);

void main(int argc, char*argv[]) {
    
    int i, j, m1;
    unsigned time;
    int c;
    extern char *optarg;
    
    while((c=getopt(argc, argv, "p:i:h")) != -1)
       {
         switch(c) {
           case 'p': numProcs=atoi(optarg);
                     if(numProcs<1) {
                       printf("P must be >=1\n");
                       exit(-1);
                     };
                     if(log_2(numProcs) == -1) {
                       printf("P must be a power of 2\n");
                       exit(-1);
                     };          	
                     break;       
           
           case 'i': I=atoi(optarg);
                     if(I < 1) {
                       printf("I must be >=1\n");
                       exit(-1);
                     }
                     break;
                     
           case 'h': printf("Usage pc <options>\n\n");
                     printf("options:\n");
                     printf("    -pP: P = number of processors; must be a power of 2.\n");
                     printf("    -nI: N = I iteration.\n");
                     printf("    -h : Print out command line options.\n");
                     printf("\nDefault: pc -p%d -n%d\n", DEFAULT_P, DEFAULT_I);   
                     exit(0);
                     break;
        }
    }

    /* Initialize the Index */
    Index = 0;

    /* Initialize the barriers and the lock */
    LOCKINIT(indexLock) 
    BARINIT(b1)   
        
    CLOCK(time);

    /* Create the consumer processes. */
    for (i = 0; i < numProcs-1; i++)
        CREATE(ConsumerStart)
        
    /* Make the master do slave work so we don't waste a processor */
    ProducerStart();        

    /* wait for all the slaves to complete */
    WAIT_FOR_END(numProcs-1)

    CLOCK(time);
    printf("Elapsed time (in cycles): %d\n", time); 
         
    printf("A(indexLock)=%x, A(Index)=%x, A(b1)=%x, A(Var)=%x\n", &indexLock, &Index, &b1, &Var);

    /*
    for (i=0; i<M; i++)
      for (j=0; j<N; j++)
         printf("ADDR(grid[%d][%d])=%x\tADDR(scratch[%d][%d])=%x\n",i,j,&grid[i][j],i,j,&scratch[i][j]);
    */
       
    }


/*====================== SlaveStart =====================================*/
/* This is the routine that each processor will be executing in parallel */
void ProducerStart() {
    int myIndex, i;
    

    LOCK(indexLock);        /* enter the critical section */    
    myIndex = Index;        /* read your ID */
    ++Index;                /* increment it, so the next will operate on ID+1 */
    UNLOCK(indexLock);      /* leave the critical section */
      
    /* the main body of a thread */
    i=0;
    while (i<I) {   
        i=i+1; 
          Var.data1 = (double) rand();
          Var.data2 = (double) rand();
          WRBACK(Var);
        BARRIER(b1, numProcs);
        WRBACK(b1);
    } /* end while*/

 }

void ConsumerStart() {
    int myIndex, i;
    Item myVar;

    LOCK(indexLock);        /* enter the critical section */    
    myIndex = Index;        /* read your ID */
    ++Index;                /* increment it, so the next will operate on ID+1 */
    UNLOCK(indexLock);      /* leave the critical section */
      
    /* the main body of a thread */
    i=0;
    OWL(Var);
    OWH(b1);
    OWL(b1);
    OWH(Index);
    while (i<I) {   
        i+=1;
        BARRIER(b1, numProcs);
        WRBACK(b1);
        myVar = Var;
    } /* end while*/

 }

/*=============== log_2 function ===============================*/
int log_2(int number)

{
  int cumulative = 1;
  int out = 0;
  int done = 0;

  while ((cumulative < number) && (!done) && (out < 50)) {
    if (cumulative == number) {
      done = 1;
    } else {
      cumulative = cumulative * 2;
      out ++;
    }
  }

  if (cumulative == number) {
    return(out);
  } else {
    return(-1);
  }
}
