#ifndef _FUNCDECL_H_
#define _FUNCDECL_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Header: funcdecl.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Kernel function declarations. Mainly applications call these,
 * but some are called by other modules of the kernel as well.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdlib.h>      /* we need (all) this just for size_t definition */
#include "timing.h"      /* definition of Thread_Time */
#include "preferences.h" /* definitions of bitsADDR and bitsLOCK */

/* 
 * kernel.c functions 
 */
    /* create a thread: */
void sys_create_thread( void (* ThreadFunction)() );
    /* wait for all the threads to end: */
void sys_wait_for_end();
    /* use this instead of exit() so the cleanup job can be done properly: */
void sys_simulation_end(int exitcode);
    /* dump statistics about the system: */
void sys_dump_statistics();
    /* get current time for the current thread, in cycles: */
Thread_Time sys_get_thread_time();
    /* get current thread's ID: */  
int sys_get_thread_ID();
    /* increment current thread's time by elapsed_thread_time cycles: */
void sys_update_thread_time(Thread_Time elapsed_thread_time);
    /* the function to reschedule the most eligible thread for execution/simulation */
void sys_scheduler();
    /* BLOCKING support: puts current thread to BLOCKED state and enters the scheduler */
void sys_thread_sleep();
    /* BLOCKING support: sets the given thread's state to READY, but does not reschedule it immediately */
void sys_thread_wakeup(int thread_to_wakeup_ID);

/* 
 * interface.c functions 
 */
void sys_memory_interface(enum Memory_Request_Type request_type, 
    size_t total_size, void *virtual_addr);
void sys_memory_loadstore(enum Memory_Request_Type request_type, 
    void *virtual_addr, size_t total_size, int isshared);
void sys_memory_lock(enum Memory_Request_Type request_type, void *lock_addr);
/* interface.c added function for prefetch*/
void sys_memory_pref(enum Memory_Request_Type request_type, void *virtual_addr, 
    int isshared);
/* Barriers */
void sys_memory_barrier_enter(enum Global_Synchro_Type global_type, void *addr, int total);
void sys_memory_barrier_in(enum Global_Synchro_Type global_type, int order);
void sys_memory_barrier_wait(enum Global_Synchro_Type global_type);
void sys_memory_barrier_out(enum Global_Synchro_Type global_type, int order);
void sys_memory_barrier_exit(enum Global_Synchro_Type global_type);

/* 
 * debug.c functions 
 */
void sys_print_source_line();
void sys_debug_source_toggle(int on);
void sys_debug_events_toggle(int on);
void sys_debug_memsim_toggle(int on);

/* 
 * auxiliary functions (from utilities.h)
 */
    /* read an item from the INI file or from the command line argument vector */
int sys_INI_read(char *key_string, int default_value);
    /* allocate system memory */
void * sys_malloc(size_t size);
    /* free some of the system memory */
void   sys_free(void *addr);

/* 
 * tracer.c functions for handling locks
 */
bitsLOCK read_lock (bitsADDR lock_addr);
void write_lock (bitsADDR lock_addr, bitsLOCK lock_value);

#endif  /* _FUNCDECL_H_ */
