#ifndef _MACROS_H_
#define _MACROS_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Header: macros.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This is what the parallel application includes; mainly ANL macros are
 * defined here, but there are also some redirections to standard functions,
 * like main() and exit(). While other simulators usually use m4 for macro 
 * substitution, a simple #include with #defines is prefered here.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "types.h"
#include "timing.h"
#include "funcdecl.h"
#include "sync.h"           /* ANL synchronization macros */

/* main() is not really main, and exit() is not really exit() */
#define main(args...) application_main(args)
#define exit(exitval) sys_simulation_end(exitval) 

/* for compatibility with SPLASH2/ANL. Not needed with Limes. */
#define MAIN_ENV                                                
#define EXTERN_ENV                                              
#define MAIN_INITENV(args...)
#define MAIN_END                                                

#define WAIT_FOR_END(NUM)   {                                   \
    STOP_TIMER;                                                 \
    sys_wait_for_end();                                         \
    START_TIMER;                                                \
    }

/* create a thread to execute void f() function */
#define CREATE(FUNCTION)    {                                   \
    STOP_TIMER;                                                 \
    sys_create_thread(FUNCTION);                                \
    START_TIMER;                                                \
    }

#define CLOCK(var)  var = sys_get_thread_time();

/* 
 * Allocate memory in user shared segment. Those who simulate DSMs
 * will probably have their own G_MALLOC here.
 */
#ifndef G_MALLOC
#define G_MALLOC(size, node...) malloc(size);
#define G_FREE(ptr) free(ptr);
#endif

/*
 * Debugging macros.
 */
#define DEBUG_SOURCE_ON  {AUG_OFF sys_debug_source_toggle(1); AUG_ON }
#define DEBUG_SOURCE_OFF {AUG_OFF sys_debug_source_toggle(0); AUG_ON }
#define DEBUG_EVENTS_ON  {AUG_OFF sys_debug_events_toggle(1); AUG_ON }
#define DEBUG_EVENTS_OFF {AUG_OFF sys_debug_events_toggle(0); AUG_ON }
#define DEBUG_MEMSIM_ON  {AUG_OFF sys_debug_memsim_toggle(1); AUG_ON }
#define DEBUG_MEMSIM_OFF {AUG_OFF sys_debug_memsim_toggle(0); AUG_ON }

/*
 * Misc macros
 */
#define getpid()        sys_get_thread_ID()

#endif