#ifndef _TIMING_H_
#define _TIMING_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Header: timing.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This defines the variable where the number of instructions for each
 * basic block is stored, and the macros used for activating/deactivating
 * this counter.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

typedef unsigned int Thread_Time;   /* change it to unsigned long long if too short for your lasting simulations! */
/*
 * sys_IC is the Instruction_Counter, but it is actually a timer, containing
 * the number of cycles the thread has spent in user mode since it last
 * leaved the kernel mode. The aug tool, in the compilation phase, inserts
 * instructions in the application assembly code to update this value
 * properly. You can analyze some of the .S files to see how it works.
 */
extern Thread_Time sys_IC;


/* 
 * Code between AUG_OFF and AUG_ON will not be augmented, nor the memory
 * references will be passed through the simulator.
 * Beware if caches are simulated: if you access a shared variable within
 * an unaugmented code, its up-to-date value may reside in caches only
 * and not in real memory!
 * Generally, AUG_OFF/ON should be used to do a printf or something without
 * disturbing process' times. If you do need access to a shared var within
 * between AUG_OFF and AUG_ON and if you simulate caches, copy the shared
 * var into a local one first, and then use the local value. 
 * Note: AUG_OFF/ON can even be outside of any function (useful if you
 * want the whole functions unaugmented).
 */
#define AUG_ON  asm("# AUG_MSG: DO_AUG");
#define AUG_OFF asm("# AUG_MSG: DONT_AUG");

/* activate timer (zero start) */
#define START_TIMER     { sys_IC = 0; AUG_ON }

/* stop the timer and update individual time by the measured value (int) */
#define STOP_TIMER      { AUG_OFF; sys_update_thread_time(sys_IC); }

/* 
 * The term "timer" remained from the time where the instructions were
 * counted through the trap mechanism. That approach did have certain
 * advantages, but was generally slow, apart from requiring Linux kernel
 * modification.
 */

#endif  /* _TIMING_H_ */