#ifndef _TYPES_H_
#define _TYPES_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Header: types.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This is where the types of request issued to the memory simulator
 * are defined. You can add your own here (for example, REQUEST_CACHE_FLUSH).
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

enum Memory_Request_Type {  /* possible request from the memory simulator */
    REQUEST_NONE    = -1,   /* no request */
    REQUEST_WRITE   = 0,    /* memory write request */
    REQUEST_READ    = 1,    /* memory read request */
    REQUEST_LOCK    = 2,    /* request for a lock (test & set) */
    REQUEST_UNLOCK  = 3,    /* unlock request (simple clear) */
    /* add your own here, if you have them. */
    REQUEST_PREF    = 4,    /* prefetch request */
    REQUEST_PREFEX  = 5,    /* prefetch-exclusive request */
    REQUEST_LPREF   = 6,    /* lazy prefetch request */
    REQUEST_WRBACK  = 7,    /* write-back request */
    REQUEST_OWL     = 8,    /* open window: defines lowest address */
    REQUEST_OWH     = 9,    /* open window: defines highest adress */
    REQUEST_CWL     = 10    /* close window: defines lowest address */ 
};
    
/* 
 * Global synchronization operations are those that involve more then just a
 * thread currently running 
 */
enum Global_Synchro_Type {  
    GLOBAL_NONE          = -1,   /* no request */
    GLOBAL_CREATED       = 0,    /* thread is created by a parent thread */
    GLOBAL_BARRIER_ENTER = 1,    /* enter barrier */
    GLOBAL_BARRIER_IN    = 2,    /* increase number of threads in barrier */
    GLOBAL_BARRIER_WAIT  = 3,    /* wait until all threads come */
    GLOBAL_BARRIER_OUT   = 4,    /* decrease number of threads in barrier */
    GLOBAL_BARRIER_EXIT  = 5,    /* exit barrier */
    /* add your own here, if you have them. */
};

/* The name of the routine called from the augmented assembly code. */
#define REQUEST_INTERFACE_ROUTINE  "_sys_memory_interface"  /* read/write/lock */
#define STRINGOP_INTERFACE_ROUTINE "_sys_memory_stringop"   /* rep movsl */

/* Others could be defined here, too. */

/*-----------------------------------------------------------------------*/
/*
 * Artificial machine instructions. Add your own...
 */
#define INSTRUCTION_LOCK    "lock"      /* names */
#define INSTRUCTION_UNLOCK  "unlock"
#define INSTRUCTION_PREF    "pref"
#define INSTRUCTION_PREFEX  "prefex"
#define INSTRUCTION_LPREF   "lpref"
#define INSTRUCTION_WRBACK  "wrback"
#define INSTRUCTION_OWL     "owl"
#define INSTRUCTION_OWH     "owh"
#define INSTRUCTION_CWL     "cwl"

enum Artificial_Instruction {           /* enums */
    ARTIF_LOCK,
    ARTIF_UNLOCK,
    ARTIF_PREF,
    ARTIF_PREFEX,
    ARTIF_LPREF,
    ARTIF_WRBACK,
    ARTIF_OWL,
    ARTIF_OWH,
    ARTIF_CWL
    };

/* 
 * Actually, an i486+ does have "LOCK", but not in the sense defined here.
 * Here, a LOCK means "acquire a lock at the address <addr>", not
 * "lock the bus". LOCKs could also have been interfaced by macros,
 * but are left as an example of creating own instructions.
 */

#endif _TYPES_H_