#ifndef _TYPES_H_
#define _TYPES_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Header: types.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This is where the types of request issued to the memory simulator
 * are defined. You can add your own here (for example, REQUEST_CACHE_FLUSH).
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

enum Memory_Request_Type {  /* possible request from the memory simulator */
    REQUEST_NONE    = -1,   /* no request */
    REQUEST_WRITE   = 0,    /* memory write request */
    REQUEST_READ    = 1,    /* memory read request */
    REQUEST_LOCK    = 2,    /* request for a lock (test & set) */
    REQUEST_UNLOCK  = 3     /* unlock request (simple clear) */
    /* add your own here, if you have them. */
    };

/* The name of the routine called from the augmented assembly code. */
#define REQUEST_INTERFACE_ROUTINE  "_sys_memory_interface"  /* read/write/lock */
#define STRINGOP_INTERFACE_ROUTINE "_sys_memory_stringop"   /* rep movsl */

/* Others could be defined here, too. */

/*-----------------------------------------------------------------------*/
/*
 * Artificial machine instructions. Add your own...
 */
#define INSTRUCTION_LOCK    "lock"      /* names */
#define INSTRUCTION_UNLOCK  "unlock"
enum Artificial_Instruction {           /* enums */
    ARTIF_LOCK,
    ARTIF_UNLOCK
    };

/* 
 * Actually, an i486+ does have "LOCK", but not in the sense defined here.
 * Here, a LOCK means "acquire a lock at the address <addr>", not
 * "lock the bus". LOCKs could also have been interfaced by macros,
 * but are left as an example of creating own instructions.
 */

#endif _TYPES_H_