/*
 *  Module: utilities.h
 *
 *  Here are definitions of some utility functions, like a function for 
 *  trace producing and initialization file support.
 */ 


/*
 *  Trace producer. If you don't like the output format, you can change it.
 */
void produce_trace (
    char *TRACEfilename,
    union Trace_file TRACEfile,
    int Trace_file_type,
    Thread_Time time_now)
{
    int i;
    char request_exists = 0;
    short int next;
    int parent;
    int create_time;
    char proc;
    char request;
    bitsADDR address;
    char size;
    char area;
    char global;
    char order;
    char total;
    char events[1000];
    events[0]='\0';

    switch (Trace_file_type) {
        case 1:  /* text trace file */    
            for (i=0; i<sys_thread_count; ++i)
                if (sys_threads[i].memory_request.requested_now) {
                    address = sys_threads[i].memory_request.virtual_addr;
                    size = (char) sys_threads[i].memory_request.size;
                    area = sys_threads[i].memory_request.area;
                    switch (sys_threads[i].memory_request.type) {
                        case REQUEST_READ  : request = 'R'; break;
                        case REQUEST_WRITE : request = 'W'; break;
                        case REQUEST_LOCK  : request = 'L'; break;
                        case REQUEST_UNLOCK: request = 'U'; break;
                        case REQUEST_PREF  : request = 'P'; break;
                        case REQUEST_PREFEX: request = 'X'; break;
                        case REQUEST_LPREF : request = 'F'; break;
                        case REQUEST_WRBACK: request = 'B'; break;
                        case REQUEST_OWL   : request = 'O'; break;
                        case REQUEST_OWH   : request = 'H'; break;
                        case REQUEST_CWL   : request = 'C'; break;
                        default: 
                            fprintf (stderr, "\nError - Unknown request type: %d\n",sys_threads[i].memory_request.type);
                            exit (1);
                    }
                    if ((!request_exists) && (time_now>=100000))
                        sprintf(events+strlen(events),"\t");
                    else
                        sprintf(events+strlen(events),"\t\t");
                    sprintf(events+strlen(events),"P%d:%c%c %u,%u",
                            i,request,area,address,size);
                    switch (sys_threads[i].global_operation.type) {
                        case GLOBAL_CREATED: 
                             sprintf(events+strlen(events),"  CREATED(P%d,T=%d)\n",
                                     sys_threads[i].global_operation.params.created.parent,
                                     sys_threads[i].global_operation.params.created.time);
                             break;
                        case GLOBAL_BARRIER_ENTER: 
                             sprintf(events+strlen(events),"  BARRIER_ENTER(%u,%d)\n",
                                     sys_threads[i].global_operation.params.barrier.virtual_addr,
                                     sys_threads[i].global_operation.params.barrier.total);
                             break;
                        case GLOBAL_BARRIER_IN:
                             sprintf(events+strlen(events),"  BARRIER_IN(%u,%d/%d)\n",
                                     sys_threads[i].global_operation.params.barrier.virtual_addr,
                                     sys_threads[i].global_operation.params.barrier.order,
                                     sys_threads[i].global_operation.params.barrier.total);
                             break;
                        case GLOBAL_BARRIER_WAIT: 
                             sprintf(events+strlen(events),"  BARRIER_WAIT(%u,%d)\n",
                                     sys_threads[i].global_operation.params.barrier.virtual_addr,
                                     sys_threads[i].global_operation.params.barrier.total);
                             break;
                        case GLOBAL_BARRIER_OUT:
                             sprintf(events+strlen(events),"  BARRIER_OUT(%u,%d/%d)\n",
                                     sys_threads[i].global_operation.params.barrier.virtual_addr,
                                     sys_threads[i].global_operation.params.barrier.order,
                                     sys_threads[i].global_operation.params.barrier.total);
                             break;
                        case GLOBAL_BARRIER_EXIT: 
                             sprintf(events+strlen(events),"  BARRIER_EXIT(%u,%d)\n",
                                     sys_threads[i].global_operation.params.barrier.virtual_addr,
                                     sys_threads[i].global_operation.params.barrier.total);
                             break;
                        case GLOBAL_NONE: 
                             sprintf(events+strlen(events),"\n");
                             break;
                        default: 
                             fprintf (stderr, "\nError - Unknown global operation type: %d\n",sys_threads[i].global_operation.type);
                             exit (1);                             
                    }
                    request_exists = 1;
                }
            if (request_exists)
                if (fprintf(TRACEfile.txt,"T=%d%s", time_now, events) < 0) {
                    fprintf (stderr, "\nError writing to trace file %s\n", TRACEfilename);
                    exit (1);
                }
            break;
        case 2:  /* binary trace file */
            for (i=0; i<sys_thread_count; ++i)
                if (sys_threads[i].memory_request.requested_now) {
                    if (!request_exists) {
                        request_exists = 1;
                        /* take place in file for the next event pointer that will be updated later */
                        next = 0;
                        if (write(TRACEfile.bin, &next, sizeof(next)) == -1) {
                            fprintf (stderr, "\nError writing to trace file %s\n", TRACEfilename);
                            exit (1);
                        }
                        next += sizeof(next);
                        /* write down the time of the event */
                        if (write(TRACEfile.bin, &time_now, sizeof(time_now)) == -1) {
                            fprintf (stderr, "\nError writing to trace file %s\n", TRACEfilename);
                            exit (1);
                        }
                        next += sizeof(time_now);
                    }
                    proc = (char) i;
                    request = (char) sys_threads[i].memory_request.type;
                    area = sys_threads[i].memory_request.area;
                    address = sys_threads[i].memory_request.virtual_addr;
                    size = (char) sys_threads[i].memory_request.size;
                    global = (char) sys_threads[i].global_operation.type;
                    if ((write(TRACEfile.bin, &proc, sizeof(proc)) == -1) ||
                        (write(TRACEfile.bin, &request, sizeof(request)) == -1) ||
                        (write(TRACEfile.bin, &area, sizeof(area)) == -1) ||
                        (write(TRACEfile.bin, &address, sizeof(address)) == -1) ||
                        (write(TRACEfile.bin, &size, sizeof(size)) == -1) ||
                        (write(TRACEfile.bin, &global, sizeof(global)) == -1))
                    {
                        fprintf (stderr, "\nError writing to trace file %s\n", TRACEfilename);
                        exit (1);
                    }
                    next = next+sizeof(proc)+sizeof(request)+sizeof(area)+sizeof(address)+sizeof(size)+sizeof(global);
                    switch (sys_threads[i].global_operation.type) {
                        case GLOBAL_CREATED: 
                             parent = sys_threads[i].global_operation.params.created.parent;
                             create_time = (int) sys_threads[i].global_operation.params.created.time;
                             if ((write(TRACEfile.bin, &parent, sizeof(parent)) == -1) ||
                                 (write(TRACEfile.bin, &create_time, sizeof(create_time)) == -1))
                             {
                                 fprintf (stderr, "\nError writing to trace file %s\n", TRACEfilename);
                                 exit (1);
                             }
                             next = next+sizeof(parent)+sizeof(create_time);
                             break;
                        case GLOBAL_BARRIER_ENTER: 
                             address = sys_threads[i].global_operation.params.barrier.virtual_addr;
                             total = sys_threads[i].global_operation.params.barrier.total;
                             if ((write(TRACEfile.bin, &address, sizeof(address)) == -1) ||
                                 (write(TRACEfile.bin, &total, sizeof(total)) == -1))
                             {
                                 fprintf (stderr, "\nError writing to trace file %s\n", TRACEfilename);
                                 exit (1);
                             }
                             next = next+sizeof(address)+sizeof(total);
                             break;
                        case GLOBAL_BARRIER_IN:
                             address = sys_threads[i].global_operation.params.barrier.virtual_addr;
                             total = sys_threads[i].global_operation.params.barrier.total;
                             order = sys_threads[i].global_operation.params.barrier.order;
                             if ((write(TRACEfile.bin, &address, sizeof(address)) == -1) ||
                                 (write(TRACEfile.bin, &order, sizeof(order)) == -1) ||
                                 (write(TRACEfile.bin, &total, sizeof(total)) == -1))
                             {
                                 fprintf (stderr, "\nError writing to trace file %s\n", TRACEfilename);
                                 exit (1);
                             }
                             next = next+sizeof(address)+sizeof(order)+sizeof(total);
                             break;
                        case GLOBAL_BARRIER_WAIT: 
                             address = sys_threads[i].global_operation.params.barrier.virtual_addr;
                             total = sys_threads[i].global_operation.params.barrier.total;
                             if ((write(TRACEfile.bin, &address, sizeof(address)) == -1) ||
                                 (write(TRACEfile.bin, &total, sizeof(total)) == -1))
                             {
                                 fprintf (stderr, "\nError writing to trace file %s\n", TRACEfilename);
                                 exit (1);
                             }
                             next = next+sizeof(address)+sizeof(total);
                             break;
                        case GLOBAL_BARRIER_OUT:
                             address = sys_threads[i].global_operation.params.barrier.virtual_addr;
                             total = sys_threads[i].global_operation.params.barrier.total;
                             order = sys_threads[i].global_operation.params.barrier.order;
                             if ((write(TRACEfile.bin, &address, sizeof(address)) == -1) ||
                                 (write(TRACEfile.bin, &order, sizeof(order)) == -1) ||
                                 (write(TRACEfile.bin, &total, sizeof(total)) == -1))
                             {
                                 fprintf (stderr, "\nError writing to trace file %s\n", TRACEfilename);
                                 exit (1);
                             }
                             next = next+sizeof(address)+sizeof(order)+sizeof(total);
                             break;
                        case GLOBAL_BARRIER_EXIT: 
                             address = sys_threads[i].global_operation.params.barrier.virtual_addr;
                             total = sys_threads[i].global_operation.params.barrier.total;
                             if ((write(TRACEfile.bin, &address, sizeof(address)) == -1) ||
                                 (write(TRACEfile.bin, &total, sizeof(total)) == -1))
                             {
                                 fprintf (stderr, "\nError writing to trace file %s\n", TRACEfilename);
                                 exit (1);
                             }
                             next = next+sizeof(address)+sizeof(total);
                             break;
                        case GLOBAL_NONE: 
                             break;
                        default:
                             fprintf (stderr, "\nError - Unknown global operation type: %d\n",sys_threads[i].global_operation.type);
                             exit (1);
                    }
                }
            if (request_exists) {
                lseek (TRACEfile.bin, -next, 1);  /* return to update the next event position */
                if (write(TRACEfile.bin, &next, sizeof(next)) == -1) {
                    fprintf (stderr, "\nError writing to trace file %s\n", TRACEfilename);
                    exit (1);
                }
                lseek (TRACEfile.bin, 0, 2);  /* go to the end of file again */
            }
            break;           
        default:  /* it should never come to this point anyway */
            fprintf (stderr, "\nError - Incorrect trace file type: %d\n", Trace_file_type);
            break;
    }
}


/*
 * sys_malloc(), a replacement for malloc(), for use with kernel and memsim,
 * intended to separate system heap from the application heap. It has two
 * major flaws: every malloc is page aligned, which means it takes a multiple
 * of 4K for each call; and, it does not implement free(). It is mostly
 * called by the memsim, in the initialization phase, when it has to allocate
 * storage for cache tags. But it really doesn't matter if an ordinary malloc()
 * is used instead, it's just a bit more neat this way.
 */
#ifdef USE_MMAP_FOR_SYS_MALLOC
#include <sys/mman.h>

void * sys_malloc(size_t size) {
    static unsigned last_malloc_addr = DATA_SEGMENT_MAXSIZE;
    caddr_t mmapped_addr;   unsigned page_aligned_size;
    #define LINE_SIZE 32    /* the CPU line size */
    #define PAGE_SIZE 4096  /* this is unlikely to ever change ... */
    if (size%LINE_SIZE !=0) size = size - size%LINE_SIZE + LINE_SIZE;   /* align at line boundary, for performance */
    if (last_malloc_addr % PAGE_SIZE != 0)  /* if addr is not page aligned */
        if (size <= PAGE_SIZE - last_malloc_addr%PAGE_SIZE) {
            void *return_addr = (void *)last_malloc_addr;
            last_malloc_addr += size;   /* small requested chunk can fit into the remainder of the last allocated page */
            return return_addr;
        }
        else    /* align the address to the beginning of the next page */
            last_malloc_addr = last_malloc_addr - last_malloc_addr%PAGE_SIZE + PAGE_SIZE;
    if ((page_aligned_size = size)%PAGE_SIZE !=0)
        page_aligned_size = size - size%PAGE_SIZE + PAGE_SIZE;
    mmapped_addr = mmap((caddr_t) last_malloc_addr, page_aligned_size,
        PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANON | MAP_FIXED, -1, 0);
    if (mmapped_addr == (caddr_t)-1) {
        fprintf(stderr, "sys_malloc: Error! mmap() failed! Using malloc() instead.\n");
        fprintf(stderr, "#undef USE_MMAP_FOR_SYS_MALLOC in  kernel.h \nif you want to avoid this message.\n");
        mmapped_addr = (caddr_t) malloc(size);
    }
    else
        last_malloc_addr += size;
    return (void *)mmapped_addr;
}
void sys_free(void *addr)       {   /* some day ... */  }
#else
void * sys_malloc(size_t size)  {   return malloc(size);    }
void sys_free(void *addr)       {   free(addr);             }
#endif USE_MMAP_FOR_SYS_MALLOC


/*
 * INI file support: this will save a lot of time if you want to vary
 * simulation parameters. Here is how to use it: call it with a string
 * that will be a key for searching the INI, and a default int value to be
 * returned in case the key does not exist in the INI, or there is no INI.
 * For example: cache_size_in_kb = sys_INI_read("cache_size",8);
 * If there exists "limes.ini" file (or some other, see '-i' kernel option),
 * with line    cache_size=16   in it, or if -dcache_size=<value> argument
 * is specified at command line (after '--'), sys_INI_read() will return
 * the specified value. If not, it will return the default given value, 8.
 * Note that command line definitions have precedence over those in the INI.
 * sys_INI_read() works only with integers.
 */
int sys_INI_read(char *key, int default_value) {
    static int first_time=1, opened=0;
    static char * INIbuf; char **argv = startup_argv;
    int retval = default_value;
    if (first_time) {
        int fh;
        struct stat statbuf;
        if (stat(INIfile, &statbuf) != -1) {
            INIbuf = (char *) sys_malloc(statbuf.st_size);
            fh = open(INIfile, 0);
            if (read(fh, INIbuf, statbuf.st_size) != -1) opened = 1;
            close(fh);
        }
        first_time = 0;
    }
    if (opened) {
        char *str = strstr(INIbuf, key);    /* format: key_string=key_value */
        if (str) sscanf(str + strlen(key)+1, "%d", &retval);
    }
    if (argv) {
        int i;  /* for argv[i] = "-dKEY=VALUE" */
        for(i=0; argv[i]; ++i)
            if (!strncmp(key, argv[i]+2, strlen(key)))
                sscanf(argv[i]+2+strlen(key)+1, "%d", &retval);
    }
    return retval;
}
