/*   
 *  +-----------------------------------------------------------+
 *  |         memsim example for simulating an ideal SMP        |
 *  +-----------------------------------------------------------+
 *  
 *  This simulator simulates an ideal memory system in the sense
 *  that all the read/write/unlock requests are served with
 *  zero latency. For a lock, it is either served immediately
 *  if the lock is free, or the PROCESSOR_LOCK_BUSY response is
 *  returned. It is up to the kernel to decide what happens with
 *  a busy lock, but it is most reasonable to select 
 *  BUSY_LOCK_SLEEP_TIME=1, BUSY_LOCK_OPTIMIZATION=1 (see kernel/preferences.h)
 *  since it's very likely that this kind of memsim will be used
 *  for parallel algorithms evaluation.
 */

#include <stdio.h>
#include <string.h>

#include "memsim_interface.h"


/*
 * Init: initialize variables and tell who will take care about the
 * real memory
 */
 
enum Memsim_Memory_Strategy memsim_init() { /* init your data structures */
    return Kernel_Maintains_Memory_Consistency;
    }
void memsim_cleanup() {puts("\x1b[2J\x1b[0mHope you've enjoyed it!");}
void memsim_fork(int parent_ID, int child_ID) {} /* on a non-DSM, nothing has to be done upon a fork */
void memsim_dump_state() {}     /* no states for this simple simulator */
void memsim_debug_toggle(int on) {} 

Thread_Time initial_thread_time(Thread_Time parent_time, int thread_number) { 
    return parent_time;         /* the time the child will have when it's born */
    }


static void gotoxy(int proc_id, int xx, int yy, char c, int inverse) {
        int x,y;
        char curstr[] = "\x1b[                        "; char *color;
        int cp       = 2;
        x = xx, y =yy;
        if (!xx || !yy) return;
        if( y>=10 ) { curstr[cp++]=y/10+'0'; y = y - 10*(y/10); }
        curstr[cp++]=y+'0';
        curstr[cp++]=';';
        if( x>=10 ) { curstr[cp++]=x/10+'0'; x = x - 10*(x/10); }
        curstr[cp++]=x+'0';
        curstr[cp++]='H';
        curstr[cp] = 0;
        if (inverse) {c=' ';strcat(curstr, "\x1b[7m");  /* inverse */}
        switch(proc_id) {
            case 0: color = "\x1b[1;37m";   break;
            case 1: color = "\x1b[1;31m";   break;
            case 2: color = "\x1b[1;32m";   break;
            case 3: color = "\x1b[1;34m";   break;
            case 4: color = "\x1b[1;36m";   break;
            case 5: color = "\x1b[1;35m";   break;
            case 6: color = "\x1b[1;33m";   break;
            default:color = "\x1b[33m";     break;
            }
        strcat(curstr, color);
        {char *s; for (s=curstr; *s; ++s) putchar(*s);}
        putchar(c);
        puts("\x1b[0m");
        x = xx, y = yy;
        }


/* 
 * The main routine, called a zillion times per simulation. It should
 * maintain locks consistent, while all the other request are immediately
 * "satisfied".
 */
void memory_simulate(Thread_Time time_before, Thread_Time time_now) {
    register int i;
    unsigned sys_heap_end;
    static int x, lastx[MAX_CPU_COUNT], y, lasty[MAX_CPU_COUNT], last_type[MAX_CPU_COUNT];
    static int first_time = 1;
    bitsADDR virtual_addr;

    for (i=0; i<CPU_COUNT; ++i)     
        if (REQUEST_EXISTS(i)) 
            switch (REQUEST_TYPE(i))    {
                case REQUEST_LOCK:
                    if ( MEMSIM_READ_LOCK(REQUEST_VIRTUAL_ADDR(i)) == 0) {
                        MEMSIM_WRITE_LOCK(REQUEST_VIRTUAL_ADDR(i),1);
                        REQUEST_RESPOND(i, PROCESSOR_SATISFIED);
                        }
                    else
                        REQUEST_RESPOND(i, PROCESSOR_LOCK_BUSY);
                    break;
                case REQUEST_UNLOCK:
                    MEMSIM_WRITE_LOCK(REQUEST_VIRTUAL_ADDR(i),0);
                    REQUEST_RESPOND(i, PROCESSOR_SATISFIED);
                    break;
                default: 
                    virtual_addr = REQUEST_VIRTUAL_ADDR(i);
                    sys_heap_end = (bitsADDR) sbrk(0);
                    if ((bitsADDR)virtual_addr >= DATA_HEAP_START && (bitsADDR)virtual_addr < DATA_SEGMENT_MAXSIZE) {
                        int block_num = 80*23*((bitsADDR)virtual_addr-DATA_HEAP_START)/(sys_heap_end - DATA_HEAP_START);
                        y = (block_num/80);
                        x = block_num-y*80;
                        ++x; ++y;
                        if (first_time) {first_time=0; puts("\x1b[2J");}
                        if (x!= lastx[i] || y != lasty[i]) {
                            gotoxy(i, lastx[i],lasty[i],
                            ((last_type[i] == REQUEST_READ) ? '.' : '_'),0);
                            gotoxy(i, x,y,
                            ((REQUEST_TYPE(i) == REQUEST_READ) ? '.' : '_'),1);
                            }
                        lastx[i]=x, lasty[i]=y;
                        last_type[i]=REQUEST_TYPE(i);
                        }
                    REQUEST_RESPOND(i, PROCESSOR_SATISFIED);
                    break;  /* the kernel will handle everything else */
                }
    }       

