/* 
 * This program reads the given trace file, calls the parser to convert
 * it into processor requests, and calls the memory simulator.
 */

#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#define extern
#include "memsim_interface.h"
#undef extern

FILE *tracefile = stdin;
static char *INIfile = "limes.ini";     /* default ini file name */

main(int argc, char *argv[]) {
    int time=-1, lasttime = 0;
    if (argc >1) {
        tracefile = fopen(argv[1], "r");
        if (!tracefile) {
            fprintf(stderr, "Cannot open trace file %s!\n", argv[1]); exit(1);
            }
        }
    if (argc>2) INIfile = argv[2];
    yytrace_init(sm_memory_request, &time, tracefile);

    memsim_init();
    
    while(yylex()) {
        int i;
        memory_simulate(lasttime, time);
        for(i=0; i<MAX_CPU_COUNT; ++i) sm_memory_request[i].requested_now = 0;
        lasttime = time + 1;
        }

    memsim_cleanup();

    }
        
void trace_printreq(struct Memory_Request * r, int i) {
    if (!r->requested_now) return;
    printf("Proc: %d: ", i);
    switch (r->type) {
        case REQUEST_READ  : printf("READ   "); break;
        case REQUEST_WRITE : printf("WRITE  "); break;
        case REQUEST_LOCK  : printf("LOCK   "); break;
        case REQUEST_UNLOCK: printf("UNLOCK "); break;
        case REQUEST_PREF  : printf("PREF   "); break;
        case REQUEST_PREFEX: printf("PREFEX "); break;
        case REQUEST_LPREF : printf("LPREF  "); break;
        case REQUEST_WRBACK: printf("WRBACK "); break;
        case REQUEST_OWL   : printf("OWL    "); break;
        case REQUEST_OWH   : printf("OWH    "); break;
        case REQUEST_CWL   : printf("CWL    "); break; 
        default            : printf("?????? "); break;
        }
    printf("%d ", r->virtual_addr);
    if (r->type == REQUEST_WRITE) printf(", %d", r->data);
    if (r->size != 4) printf("   (size %d) ", r->size);
    printf("\n");
    }   

/*
 * INI file support 
 */

int sys_INI_read(char *key, int default_value) {
    static int first_time=1, opened=0;
    static char * INIbuf;
    int retval = default_value;
    if (first_time) {
        int fh;
        struct stat statbuf;
        if (stat(INIfile, &statbuf) != -1) {
            INIbuf = (char *) malloc(statbuf.st_size);
            fh = open(INIfile, 0);
            if (read(fh, INIbuf, statbuf.st_size) != -1) opened = 1;
            close(fh);
            }
        first_time = 0;
        }
    if (opened) {
        char *str = strstr(INIbuf, key);    /* format: key_string=key_value */
        if (str) sscanf(str + strlen(key)+1, "%d", &retval);
        }
    return retval;
    }
