/*   
 *  +-----------------------------------------------------------+
 *  |         memsim example for simulating an ideal SMP        |
 *  +-----------------------------------------------------------+
 *  
 *  This simulator simulates an ideal memory system in the sense
 *  that all the read/write/unlock requests are served with
 *  zero latency. For a lock, it is either served immediately
 *  if the lock is free, or the PROCESSOR_LOCK_BUSY response is
 *  returned. It is up to the kernel to decide what happens with
 *  a busy lock, but it is most reasonable to select 
 *  BUSY_LOCK_SLEEP_TIME=1, BUSY_LOCK_OPTIMIZATION=1 (see kernel/preferences.h)
 *  since it's very likely that this kind of memsim will be used
 *  for parallel algorithms evaluation.
 */

#include <stdio.h>
#include <string.h>

#include "memsim_interface.h"

/*
 * Init: initialize variables and tell who will take care about the
 * real memory
 */
enum Memsim_Memory_Strategy memsim_init() { /* init your data structures */
    return Kernel_Maintains_Memory_Consistency;
    }
void memsim_cleanup() {}
void memsim_fork(int parent_ID, int child_ID) {} /* on a non-DSM, nothing has to be done upon a fork */
void memsim_dump_state() {}     /* no states for this simple simulator */
void memsim_debug_toggle(int on) {} 

Thread_Time initial_thread_time(Thread_Time parent_time, int thread_number) { 
    return parent_time;         /* the time the child will have when it's born */
    }

/* 
 * The main routine, called a zillion times per simulation. It should
 * maintain locks consistent, while all the other request are immediately
 * "satisfied".
 */
void memory_simulate(Thread_Time time_before, Thread_Time time_now) {
    register int i;
    for (i=0; i<CPU_COUNT; ++i)     
        if (REQUEST_EXISTS(i)) 
            switch (REQUEST_TYPE(i)) {
                case REQUEST_LOCK:
                    if ( MEMSIM_READ_LOCK(REQUEST_VIRTUAL_ADDR(i)) == 0) {
                        MEMSIM_WRITE_LOCK(REQUEST_VIRTUAL_ADDR(i),1);
                        REQUEST_RESPOND(i, PROCESSOR_SATISFIED);
                        }
                    else
                        REQUEST_RESPOND(i, PROCESSOR_LOCK_BUSY);
                    break;
                case REQUEST_UNLOCK:
                    MEMSIM_WRITE_LOCK(REQUEST_VIRTUAL_ADDR(i),0);
                    REQUEST_RESPOND(i, PROCESSOR_SATISFIED);
                    break;
                default:
                    REQUEST_RESPOND(i, PROCESSOR_SATISFIED);
                    break;
            }
        else
            REQUEST_RESPOND(i, PROCESSOR_NORESPONSE);
}       

