#ifndef _IAT_H_
#define _IAT_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: iat.h
 * Author: Aleksandar Milenkovic, emilenka@etf.bg.ac.yu
 *
 * Defines the Injection Anticipation Table
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "snoopy.h"

// default number of entries
const DEFAULT_IAT_SIZE = 16;   

class IAT {
  protected:
    // num. of entries, adjustable at init time
    int iat_size; //number of entries
    // write pointers
    int lwp; // last write pointer
 
    struct Entry {
        bitsADDR laddr;  //the lowest address
        bitsADDR haddr;  //the highest address
        bool valid; // valid=1: request is valid 
        Entry() {laddr=haddr=(bitsADDR)-1; valid=0;}
        } *table;
  public:
    void Initialize(int iat_size=DEFAULT_IAT_SIZE); 
    void Introduce();
    bool isvalid(bitsADDR addr);
    void openWinLAddr(bitsADDR addr);
    void openWinHAddr(bitsADDR addr);
    void closeWin(bitsADDR addr);
    void dump_state(int index);
    };


#ifdef PRAGMA_IMPLEMENTATION_IAT
/////////////////////////////////////////////////////--- IMPLEMENTATION

#include <stdio.h>      // needed for the dump

// Initialize method must be called before the IAT is used.

void IAT::Initialize(int default_iat_size) {   

    // initialize cache geometry
    iat_size = sys_INI_read("IAT_size", default_iat_size);

    // instantiate new entries; they will be automatically initialized 
    table = new Entry[iat_size];   
    // set pointers
    lwp=-1;
    // set flags
}       

// print buffer geometry to stdout

void IAT::Introduce() {
    PRINTF("IAT - size: %d \n", iat_size); 
}

// look up buffer for given address
// this will make to perform snooping of relevant line???

bool IAT::isvalid(bitsADDR addr) {
    int valid=0;
    for(int i=0; i<iat_size; i++)
      if((addr<=table[i].haddr) && (addr>=table[i].laddr) && table[i].valid) {
        valid=1;
        break;
      }
    return valid;
}

//open the address window (init the lowest address)
void IAT::openWinLAddr(bitsADDR addr) {
   int i=0;
   //find empty entry
   while(table[i].valid) i++;
   if(i<iat_size) lwp=i;
   else //table is full
     lwp=(lwp+iat_size/2+1)%iat_size;
   table[lwp].laddr = addr;
   table[lwp].haddr = addr;
   table[lwp].valid=1;
}

// put the highest address
void IAT::openWinHAddr(bitsADDR addr) {
   table[lwp].haddr = addr;
}

// close window
void IAT::closeWin(bitsADDR addr) {
  int i=0;
  for(i=0; i<iat_size; i++) 
      if(table[i].laddr==addr) table[i].valid=0;
}  	
	
	
	

void IAT::dump_state(int index) {
    for(int i=0; i<iat_size; i++)
       if(table[i].valid)
         PRINTF("IAT%d[%d]: laddr=%x, haddr=%x, valid=%d\n", 
                index, i, table[i].laddr, table[i].haddr, table[i].valid);
}              
      

#endif PRAGMA_IMPLEMENTATION_IAT
#endif _IAT_H_
