/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: memsim.cc
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * The main (and generic) simulation class, followed by the C++ to C
 * interface to the simulation kernel.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "snoopy.h"
#include TOPOLOGY_HDR

///*
#ifdef DETACHED_MEMSIM
static int debug = 1;
#else
static int debug = 0;
#endif
//*/

//static int debug = 1;

class Memsim {
    Thread_Time last_simulation_time;
  public:
    void init() { Topology::system_init(); }
    void end()  { Topology::system_cleanup(last_simulation_time); }
    void simulate(Thread_Time time_before, Thread_Time time_now);
    void dump_state()   
        { Topology::system_dump_state(); }
    void memsim_print_requests(Thread_Time time);
    void memsim_print_responses(Thread_Time time);
    } memsim;

/*
 * The main routine. Called by the kernel every time a new request or a set
 * of requests appear, and in successive cycles where all the processors
 * are stalled.
 */
void Memsim::simulate(Thread_Time time_before, Thread_Time time_now) {

    for(Thread_Time time=time_before; time<=time_now; ++time) {
        int i;

        // optimization: if the whole system is in initial time,
        // shift current time to present
        if (time < time_now && Topology::system_isidle()) time = time_now;

        // if time now is time_now, wherein there are some requests, for
        // that is the reason we have been called in the first place,
        // collect them. The only situation where before != now is when
        // in before-1 everyone was satisfied, and we had to finish a
        // write or something. From before to now-1 there are no requests.
        if (time == time_now) {
            if (debug)  
                for(i=0; i<CPU_COUNT; ++i) // don't print anything if there are no requests
                    if (REQUEST_EXISTS(i)) { memsim_print_requests(time); break; }
            Topology::system_collect_requests();
            }
        //if (debug) dump_state();
        //dump_state();
	Topology::setCurrentTime(time);
        Topology::system_cycle();
        if (debug) { memsim_print_responses(time); FFLUSH; }
	if (debug) {
 	   COUT << "=> [ " << time << "+]\n"; 
           dump_state(); 
	   COUT << "====================================================\n\n";}
        //memsim_print_responses(time); FFLUSH;
        }
    last_simulation_time = time_now;
    }       

/*
 * These two are used for debugging.
 */
void Memsim::memsim_print_requests(Thread_Time time) {
    struct Memory_Request *req;
    char line[100]; int i;
    int pos =0;
    for(i=0; i<78; ++i) line[i] = ' ';
    sprintf(line, "==>  [%3d]:", time); pos += 15;
    #define BASE (bitsADDR) DATA_SEGMENT_START
    for(i=0; i<CPU_COUNT; ++i) {
        REQUEST_FILL(i, req);
        if (REQUEST_EXISTS(i)) 
            sprintf(line+pos, "P%d:", i); 
        pos +=4;
            
        if (REQUEST_EXISTS(i)) {
            REQUEST_FILL(i, req);

            switch (req->type) {
                case REQUEST_READ  : sprintf(line+pos, "R %x",
                    req->virtual_addr - (req->isshared ? BASE : 0)); 
                    break;
                case REQUEST_WRITE : sprintf(line+pos, "W %x",
                    req->virtual_addr - (req->isshared ? BASE : 0)); 
                    break;
                case REQUEST_LOCK  : sprintf(line+pos, "L %x",
                    req->virtual_addr - (req->isshared ? BASE : 0)); 
                    break;
                case REQUEST_UNLOCK: sprintf(line+pos, "U %x",
                    req->virtual_addr - (req->isshared ? BASE : 0)); 
                    break;
                case REQUEST_PREF:   sprintf(line+pos, "P %x",
                     req->virtual_addr - (req->isshared ? BASE : 0));
                     break;
                case REQUEST_PREFEX:   sprintf(line+pos, "X %x",
                     req->virtual_addr - (req->isshared ? BASE : 0));
                     break;
                case REQUEST_LPREF:   sprintf(line+pos, "F %x",
                     req->virtual_addr - (req->isshared ? BASE : 0));
                     break;
                case REQUEST_WRBACK:   sprintf(line+pos, "B %x",
                     req->virtual_addr - (req->isshared ? BASE : 0));
                     break;
                case REQUEST_OWL:   sprintf(line+pos, "O %x",
                     req->virtual_addr - (req->isshared ? BASE : 0));
                     break;
                case REQUEST_OWH:   sprintf(line+pos, "H %x",
                     req->virtual_addr - (req->isshared ? BASE : 0));
                     break;
                case REQUEST_CWL:   sprintf(line+pos, "C %x",
                     req->virtual_addr - (req->isshared ? BASE : 0));
                     break;

                default : break;
                }
            switch (req->size) {
                case 1 : line[pos+1] = 'b'; break;
                case 2 : line[pos+1] = 'w'; break;
                default: break;
                }
            }
        else    
        pos += 12;
        }
    for(i=0; i<78; ++i) if (!line[i]) line[i] = ' ';
    line[78] = 0;
    COUT << line << '\n';
    }

void Memsim::memsim_print_responses(Thread_Time time) {
    struct Memory_Request *req;
    char line[100];
    int pos =0;
    const DELC = '-';
    for(int i=0; i<78; ++i) line[i] = DELC;
    sprintf(line, "<==  [%3d]: ", time); pos += 15;
    for(i=0; i<CPU_COUNT; ++i) {
        REQUEST_FILL(i, req);
        switch (req->satisfied) {
            case PROCESSOR_SATISFIED: 
                sprintf(line+pos, "P%d: OK ", i);
                break;
            case PROCESSOR_LOCK_BUSY:
                sprintf(line+pos, "P%d: LOCK_BUSY ",i); break;
            case PROCESSOR_STALLED:
                sprintf(line+pos, "P%d: STALLED ",i); break;
            default : break;
            }
        pos += 16;
        }
    for(i=0; i<78; ++i) if (!line[i]) line[i] = DELC;
    line[78] = 0;
    COUT << line << "\n";
    }

/**************************************************************************/
// Redirection of memory allocation through "new" to "sys_malloc()":
// that separates memsim's heap from the application heap. Not really
// necessary, but can't hurt. You can leave this in your memsims.
// If you comment it out, an ordinary "operator new" will be used instead.
void * operator new(size_t size) {
    return sys_malloc(size); 
    }

/**************************************************************************/
// C++ to C interface. Guess it's common for all the memsim's.

enum Memsim_Memory_Strategy memsim_init() { 
    memsim.init(); 
    /* 
     * this means that memsim only simulates events, but the kernel
     * is responsible for actual reads, writes, locks and unlocks.
     */
    return Kernel_Maintains_Memory_Consistency;
    }

void memsim_cleanup() { 
    memsim.end(); 
    }

void memory_simulate(Thread_Time time_before, Thread_Time time_now) {
    memsim.simulate(time_before, time_now);
    }

void memsim_dump_state() {
    memsim.dump_state();
    }

Thread_Time initial_thread_time(Thread_Time parent_time, int thread_number) { 
    return parent_time; 
    }

void memsim_fork(int parent_ID, int child_ID) {
    }

void memsim_debug_toggle(int on) {
    debug = on;
    }   