#ifndef _TOPOLOGY_H_
#define _TOPOLOGY_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: topology.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Defines the methods that the other modules will use to read their
 * input ports, and certain globally important functions, like system_cycle().
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"

// declares the types of modules in the system
class CACHE_CLASS;
//class BUS_CLASS;
class IO_BUF_CLASS;

struct BusStats {
   int rdc;
   int rdxc;
   int rwbc;
   int swbc;
   int invc;
   int iwbc;
   int irpc;
};
struct ProcStats {
   int req;
   int rd_req;
   int rd_hit;
   int rd_miss;
   int wr_req;
   int wr_hit;
   int wr_miss;
   int wr_inv;
   int lo_req;
   int lo_hit;
   int lo_miss;
   int lo_inv;
   int un_req;
   int un_hit;
   int un_miss;
   int un_inv;
   int pr_req;
   int pr_hit;
   int pr_miss;
   int px_req;
   int px_hit;
   int px_miss;
   int px_inv;
};

class Base_Topology {
  //protected:        // each module is instantiated through a pointer to it
    public:
    static IO_BUF_CLASS             * io_buf[MAX_CPU_COUNT];
    static CACHE_CLASS              * cache[MAX_CPU_COUNT];
  //  static BUS_CLASS                * bus;
    static Thread_Time		      current_time;
    static BusStats		      bstats; // bus statistics
    static ProcStats		      pstats; // processor statistics
  //public:
    // services for individual types of modules
    static bool io_buf_input(int index);
    static bool cache_read_request(int index);
    //static void cache_read_bus(int index);
    //static bool bus_read_REQs();
    //static bool bus_read_LOCK_REQs();
    //static void bus_read_contents(int master_index);

    // system services. These exist for each simulator.
    static void system_init();
    static void system_cleanup(Thread_Time last_time);
    static bool system_isidle();
    static void system_dump_state();
    static void system_collect_requests();
    static void system_cycle();
    static void setCurrentTime(Thread_Time time_now) {
       	current_time = time_now;
       	}
    static Thread_Time getCurrentTime() {
        return current_time;
        }
    };

typedef TOPOLOGY_CLASS Topology;


#endif // _TOPOLOGY_H_