#ifndef _SNOOPY_H_
#define _SNOOPY_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: snoopy.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Defines macros and aliases that will be seen by all the other modules.
 * Its input argument is symbol SNOOP_TYPE, defined in the makefile that
 * initiates the compilation.
 *
 * Note: This file is extended with macros needed to support MESI protocol
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

extern "C" {
#include "memsim_interface.h"
}

// registry base :-)
#define snoop_type_wti          1
#define snoop_type_ber          2
#define snoop_type_dra          3
#define snoop_type_win          4
#define snoop_type_mesi		5

// SNOOP_TYPE is defined externally, with -D (from the makefile)
// It is used to distinguish between the header files and class names 
// that are different for each of the cache protocols.

// this definitions are common for all the caches:

#define TOPOLOGY_CLASS          Base_Topology
#define TOPOLOGY_FRIENDS        friend class Base_Topology;
#define TOPOLOGY_HDR            "topology.h"

#define BUS_HDR                 "bus.h"
#define BUS_CLASS               Bus_Module
#define IO_BUF_HDR              "io_buf.h"
#define IO_BUF_CLASS            IO_Buf

// while these differ

#if SNOOP_TYPE==snoop_type_wti  // WTI dependencies
    #define CACHE_HDR               "wti_cache.h"
    #define CACHE_CLASS             WTI_Cache
#endif //snoop_type_wti

#if SNOOP_TYPE==snoop_type_ber  // Berkeley dependencies
    #define CACHE_HDR               "ber_cache.h"
    #define CACHE_CLASS             Ber_Cache
#endif //snoop_type_ber

#if SNOOP_TYPE==snoop_type_dra  // Dragon dependencies
    #define CACHE_HDR               "dra_cache.h"
    #define CACHE_CLASS             Dra_Cache
#endif //snoop_type_dra

#if SNOOP_TYPE==snoop_type_win  // WIN dependencies
    #define CACHE_HDR               "win_cache.h"
    #define CACHE_CLASS             WIN_Cache
#endif //snoop_type_dra

#if SNOOP_TYPE==snoop_type_mesi // MESI dependecies
    #define CACHE_HDR		    "mesi_cache.h"
    #define CACHE_CLASS		     MESI_Cache
#endif  //snoop_type_mesi

// tuning the behavior of cache controllers
#if 1                       // selects whether the cache element will use
    #define LOCK_LINE_AWARE     // the bus or a special line for entering the 
#endif                      // "critical section"

#if 0
    #define DUMM_LOCK           // dumm means does test&set instead of test&test&set
#endif

#endif _SNOOPY_H_