#ifndef _ABUFFER_H_
#define _ABUFFER_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: abuffer.h
 * Author: Aleksandar Milenkovic, emilenka@etf.bg.ac.yu
 *
 * Defines the generic address buffer 
 * (needed for write-buffer,prefetch-buffer, etc)
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "snoopy.h"

// default number of entries
const DEFAULT_NENTRIES = 16;   
enum BuffR_Type {NOTYPE, PREF, PREFEX};

class ABuffer {
  protected:
    // num. of entries, adjustable at init time
    int nentries;
    // read/write pointers
    int rp; // read pointer
    int wp; // write pointer
    int full; // abuffer is full
    int empty; // abuffer is empty
 
    struct Entry {
        bitsADDR adr;
        BuffR_Type rtype;
        int size;  // size of the requested data
        bool valid; // valid=1: request is valid 
        Entry() {adr=(bitsADDR)-1; rtype=NOTYPE; size=0; valid=0;}
        } *entries;
  public:
    void Initialize(int nentires=DEFAULT_NENTRIES); 
    void Introduce();
    bool isvalid(bitsADDR addr);
    bool isFull(void) {return full;}
    bool isEmpty(void) {return empty;}
    void write(bitsADDR adr, int size, BuffR_Type rtype);
    void read(bitsADDR *adr, int *size, BuffR_Type *rtype);
    void dump_state(int index);
    };


#ifdef PRAGMA_IMPLEMENTATION_ABUFFER
/////////////////////////////////////////////////////--- IMPLEMENTATION
#include <stdio.h>      // needed for the dump

// Initialize method must be called before the abuffer is used.

void ABuffer::Initialize(int default_nentries) {   

    // initialize cache geometry
    nentries = sys_INI_read("number_of_entries", default_nentries);

    // instantiate new entries; they will be automatically initialized 
    entries = new Entry[nentries];                  // by their ctors
    // set pointers
    rp=wp=0;
    // set flags
    full=0; empty=1;
    }       

// print buffer geometry to stdout

void ABuffer::Introduce() {
    PRINTF("Buffer: nentries: %d \n", nentries); 
    }

// look up buffer for given address
// this will make to perform snooping of relevant line???
bool ABuffer::isvalid(bitsADDR adr) {
    int valid=0;
    for(int i=0; i<nentries; i++)
      if((entries[i].adr==adr) && entries[i].valid) {
        valid=1;
        break;
      }
    return valid;
}


void ABuffer::write(bitsADDR adr, int size, BuffR_Type rtype) {
   entries[wp].adr = adr;
   entries[wp].rtype = rtype;
   entries[wp].size = size;
   entries[wp].valid=1;
   wp=(wp+1)%nentries;
   if(rp==wp) full=1;
   if(empty) empty=0;
}

void ABuffer::read(bitsADDR *adr, int *size, BuffR_Type *rtype) {
   *adr=entries[rp].adr;
   *rtype=entries[rp].rtype;
   *size=entries[rp].size;
   entries[rp].valid=0;
   rp=(rp+1)%nentries;
   if(rp==wp) empty=1;
   if(full) full=0;
}

void ABuffer::dump_state(int index) {
    for(int i=0; i<nentries; i++)
       if(entries[i].valid)
         PRINTF("PB%d[%d]: addr=%x, rtype=%d, size=%d, valid=%d\n", 
                index, i, entries[i].adr, entries[i].rtype, 
                entries[i].size, entries[i].valid);
}              
      

#endif PRAGMA_IMPLEMENTATION_ABUFFER
#endif _ABUFFER_H_
