#ifndef _BUS_SIGNALS_H_
#define _BUS_SIGNALS_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: bus_signals.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Definitions of the signals that can appear on the bus.
 * 
 * This file is modified by A. Milenkovic.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
/*
 * Bus cycle types for MESI protocol are given in the table below:
 * | bus cycle type  |  memcmd        | bus signal   |
 * ---------------------------------------------------
 * | INVALIDATE      | request_write  | INV          |
 * | READ-EXCLUSIVE  | request_read   | INV          |
 * | READ            | request_read   |  -           |
 * | WRITE           | request_write  |  -           |
 * ---------------------------------------------------
 */

struct Bus_Signals {
    bitsADDR            addr;           // address
    int                 size;           // size (1..DATA_WIDTH)
    Memory_Request_Type memcmd;         // READ/WRITE
    bool                SNOOP_REQ;      // SNOOP REQUEST bus line
    bool                SHARED;         // SHARED bus line
    bool                DIRTY;          // DIRTY bus line
    bool 		INV;		// INVALIDATE bus line  
    reset() {
        memcmd   = REQUEST_NONE;
        SNOOP_REQ = SHARED = DIRTY = INV = 0;
        }
    Bus_Signals() { reset(); }
    void dump()   {                     // for debugging
        char *memcmd_desc;
        switch (memcmd) {
            case REQUEST_READ  : memcmd_desc = "READ"; break;
            case REQUEST_WRITE : memcmd_desc = "WRIT"; break;
            case REQUEST_NONE  : memcmd_desc = "NONE"; break;
            default            : memcmd_desc = "????"; break;
            }
        PRINTF("A:%04x  C:%s  SNP_R:%d  SH:%d  DIRT:%d  INV:%d",
            addr, memcmd_desc, SNOOP_REQ, SHARED, DIRTY, INV);
        }
    };

#endif // _BUS_SIGNALS_H_