#ifndef _BUS_SIGNALS_H_
#define _BUS_SIGNALS_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: bus_signals.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Definitions of the signals that can appear on the bus.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"

struct Bus_Signals {
    bitsADDR            addr;           // address
    int                 size;           // size (1..DATA_WIDTH)
    Memory_Request_Type memcmd;         // READ/WRITE
    bool                SNOOP;          // SNOOP=1 activates snoopy contollers
    bool                INV;            // for inavlidating lines in caches
    bool                SHARE;          // for line shareness detection
    bool                DIRTY;          // DIRTY line is activated if block is modified
    reset() {
        memcmd = REQUEST_NONE;
        SNOOP  = INV = SHARE = DIRTY = 0;
    }
    Bus_Signals() { reset(); }
    void dump()  {                      // for debugging
        char *memcmd_desc;
        switch (memcmd) {
            case REQUEST_READ  : memcmd_desc = "READ"; break;
            case REQUEST_WRITE : memcmd_desc = "WRITE"; break;
            case REQUEST_NONE  : memcmd_desc = "NONE"; break;
            default            : memcmd_desc = "????"; break;
        }
        PRINTF("A:%04u  C:%s  SNOOP:%d  INV:%d  SHARE:%d  DIRTY:%d ",
            addr, memcmd_desc, SNOOP, INV, SHARE, DIRTY);
    }
};

#endif // _BUS_SIGNALS_H_