#ifndef _BUS_SIGNALS_H_
#define _BUS_SIGNALS_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: bus_signals.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Definitions of the signals that can appear on the bus.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"

struct Bus_Signals {
    bitsADDR            addr;           // address
    int                 size;           // size (1..DATA_WIDTH)
    Memory_Request_Type memcmd;         // READ/WRITE
    bool                SNOOP;          // SNOOP=high activates snoopy contollers
    union {
        bool            INV;            // INV for invalidate protocols,
        bool            WR_MISS;        // or WR_MISS for Dragon
        };
    bool                SUPPLY;         // high if a cache will do the supply
    bool                SH_LINE;        // for line shareness detection
    bool                WORD;           // for WIN: is it a word or a whole line
    reset() {
        memcmd   = REQUEST_NONE;
        SNOOP    = INV = SUPPLY = SH_LINE = 0;
        WORD     = 0;
        }
    Bus_Signals() { reset(); }
    void dump()   {                     // for debugging
        char *memcmd_desc;
        switch (memcmd) {
            case REQUEST_READ  : memcmd_desc = "READ"; break;
            case REQUEST_WRITE : memcmd_desc = "WRIT"; break;
            case REQUEST_NONE  : memcmd_desc = "NONE"; break;
            default            : memcmd_desc = "????"; break;
            }
        PRINTF("A:%04u  C:%s  SNP:%d  INV:%d  SUP:%d  SH:%d",
            addr, memcmd_desc, SNOOP, INV, SUPPLY, SH_LINE);
        }
    };

#endif // _BUS_SIGNALS_H_