#ifndef _TOPOLOGY_H_
#define _TOPOLOGY_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: topology.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Defines the methods that the other modules will use to read their
 * input ports, and certain globally important functions, like system_cycle().
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"

// declares the types of modules in the system
class CACHE_CLASS;
class BUS_CLASS;
class IO_BUF_CLASS;

class Base_Topology {
  protected:        // each module is instantiated through a pointer to it
    static IO_BUF_CLASS             * io_buf[MAX_CPU_COUNT];
    static CACHE_CLASS              * cache[MAX_CPU_COUNT];
    static BUS_CLASS                * bus;
    static Thread_Time		    current_time;
  public:
    // services for individual types of modules
    static bool io_buf_input(int index);
    static bool cache_read_request(int index);
    static void cache_read_bus(int index);
    static bool bus_read_REQs();
    static bool bus_read_LOCK_REQs();
    static void bus_read_contents(int master_index);

    // system services. These exist for each simulator.
    static void system_init();
    static void system_cleanup(Thread_Time last_time);
    static bool system_isidle();
    static void system_dump_state();
    static void system_collect_requests();
    static void system_cycle();
    static void setCurrentTime(Thread_Time time_now) {
         current_time = time_now;
      };
    static Thread_Time getCurrentTime() {
         return current_time;
      };
    };

typedef TOPOLOGY_CLASS Topology;


#endif // _TOPOLOGY_H_
