#ifndef _SYNC_H_
#define _SYNC_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Header: sync/abstract/sync.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This header defines declarations for certain ANL synchronization
 * primitives. It is included by "macros.h", that the application sees.
 * It is also included by "sync.c" where actual primitives are implemented.
 * Lock and barrier structures defined here are made for implementation
 * of synchronization primitives in an "abstract" way, so that the memory
 * simulator never sees a LOCK or an UNLOCK request. Instead, a thread
 * blocks on a busy LOCK, and is awaken by a thread who does UNLOCK.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "preferences.h"	/* needed to see the MAX_THREAD_COUNT constant */

/* locks */				
struct LockStruct {
	int value;				/* 0 - free, 1 - busy */
	int blocked_threads_queue[MAX_THREAD_COUNT];
	int queue_top, queue_length;
	};
void sys_lock_init(struct LockStruct *);
void sys_lock_acquire(struct LockStruct *);
void sys_lock_release(struct LockStruct *);

#define LOCKDEC(L) 		struct LockStruct L;
#define LOCKINIT(L) 	sys_lock_init(&L);
#define LOCK(L)    		sys_lock_acquire(&L);
#define UNLOCK(L) 		sys_lock_release(&L);

/* lock arrays */
#define ALOCKDEC(LA,N)	struct LockStruct LA[N];
#define ALOCKINIT(LA,N)	{int _i; for(_i=0;_i<(N);++_i) LOCKINIT(LA[_i])}
#define ALOCK(LA,I)		LOCK(LA[I]);
#define AULOCK(LA,I)	UNLOCK(LA[I]);

/* barriers */
struct BarrierStruct {
	int count;
	int blocked_threads_list[MAX_THREAD_COUNT];
	};
void sys_barrier_init(struct BarrierStruct *B);
void sys_barrier(struct BarrierStruct *B, int N);

#define BARDEC(B) 		struct BarrierStruct B;
#define BARINIT(B) 		sys_barrier_init(&B);
#define BARRIER(B,N)	sys_barrier(&B, N);

#endif _SYNC_H_