#ifndef _SYNC_H_
#define _SYNC_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Header: sync/realistic/sync.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This header defines declarations for certain ANL synchronization
 * primitives. It is included by "macros.h", that the application sees.
 * It is also included by "sync.c" where actual primitives are implemented.
 * The structures defined here are made for implementation of synchronization
 * primitives in a "realistic" manner. See sync.c for details.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* locks */				
struct LockStruct {
	int value;			/* 0 - free, 1 - busy */
	};
void sys_lock_init(struct LockStruct *);
void sys_lock_acquire(struct LockStruct *);
void sys_lock_release(struct LockStruct *);
/* added instr */
void sys_pref(bitsADDR);

#if 1					/* locks are implemented through system calls */
#define LOCKDEC(L) 		struct LockStruct L;
#define LOCKINIT(L) 	sys_lock_init(&L);
#define LOCK(L)    		sys_lock_acquire(&L);
#define UNLOCK(L) 		sys_lock_release(&L);
#define PREF(A)		        sys_pref(&A);

#else	/* or, try this: locks implemented through special asm instrs */
#define LOCKDEC(L) 		struct LockStruct L;
#define LOCKINIT(L) 	L.value = 0;
#define LOCK(L)    		asm(INSTRUCTION_LOCK   " %0" : "=m" (L.value) );
#define UNLOCK(L) 		asm(INSTRUCTION_UNLOCK " %0" : "=m" (L.value) );
#endif 

/* lock arrays */
#define ALOCKDEC(LA,N)	struct LockStruct LA[N];
#define ALOCKINIT(LA,N)	{int _i; for(_i=0;_i<(N);++_i) LOCKINIT(LA[_i])}
#define ALOCK(LA,I)		LOCK(LA[I]);
#define AULOCK(LA,I)	UNLOCK(LA[I]);

/* barriers */
struct BarrierStruct {
	LOCKDEC(lock);
	LOCKDEC(sleeplock);
	int sleepers;
	};
void sys_barrier_init(struct BarrierStruct *B);
void sys_barrier(struct BarrierStruct *B, int N);

#define BARDEC(B) 		struct BarrierStruct B;
#define BARINIT(B) 		sys_barrier_init(&B);
#define BARRIER(B,N)	sys_barrier(&B, N);

#endif _SYNC_H_