/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: string.c
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Here are (most) standard library functions related to strings.
 * They are compiled and augmented just like ordinary application
 * files. None of them is called by the kernel, only by the application.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <string.h>
#include <stdlib.h>

#include "preferences.h"
#include "sys_stdlib.h"

size_t sys_strlen(const char *s) {
    register int i=0;
    while (*s++) ++i;
    return (size_t) i;
    }
char * sys_strcpy(char *dest, const char *src) {
    register char *d = dest;
    while ((*d++ = *src++));
    return dest;
    }
char * sys_strncpy(char *dest, const char *src, size_t n) {
    register int i=0;
    while ((dest[i] = src[i]) && i<n)   ++i;
    dest[i] = '\0';
    return dest;
    }
char * sys_strcat(char *dest, const char *src) {
    return sys_strcpy(dest + sys_strlen(dest), src);
    }
char * sys_strncat(char *dest, const char *src, size_t n) {
    return sys_strncpy(dest + sys_strlen(dest), src, n);
    }
char * sys_strchr(const char *s, int c) {
    for(;;++s) 
        if ((int)*s == c) return (char *)s;
        else if (!*s)     return NULL;
    }       
int sys_strcmp(const char *s1, const char *s2) {
    register int i=0;
    if (s1 == s2) return 0;
    while (s1[i] == s2[i]) 
        if (s1[i++] == '\0') return 0;
    return s1[i] - s2[i];
    }
int sys_strncmp(const char *s1, const char *s2, size_t n) {
    register int i=0;
    if (s1 == s2) return 0;
    while (s1[i] == s2[i] && i<n) 
        if (s1[i++] == '\0') return 0;
    return s1[i] - s2[i];
    }
void * sys_memcpy(void *dest, const void *src, size_t n) {
    int offset = 0;
    while (offset+DATA_WIDTH <= n)  {
        *(bitsMAX *)(dest+offset) = *(bitsMAX *)(src+offset);
        offset += DATA_WIDTH;
        } 
    for(; offset<n; ++offset)   
        *(char *)(dest+offset) = *(char *)(src+offset);
    return dest;
    }

/*
 * Etc. Some day I might add others (God didn't create functions like
 * strstr() for nothing :-) but this does not make much sense anyway.
 * These functions are barely ever called during the parallel computation
 * phase.
 */
