/*
 * This file contains a set of functions needed for work with data structures
 * described in tracer_types.h file, as well as functions that are used to 
 * handle locks and barriers. These functions are used within tracer.c 
 * and trace.lex files.
 *
 * Author:  Igor Ikodinovic  ( igi@eunet.yu )
 *          Department of Computer Engineering
 *          Faculty of Electrotechnical Engineering
 *          University of Belgrade
 *          Serbia, Yugoslavia
 *          1998
 *
 */


/*
 *
 * Procedures for working with hash tables 
 *
 */

bitsLOCK read_from_hash_table (type_hash_table *hash_table, bitsADDR addr) {
    struct list_element_1 *curr_element;
    curr_element = hash_table[addr % HASHTABLE_SIZE].list_head;
    while (curr_element) {
        if (curr_element->address == addr) return (curr_element->value);
        curr_element = curr_element->next;
    } 
    return (0);
}

void write_to_hash_table (type_hash_table *hash_table, bitsADDR addr, bitsLOCK value) {
    struct list_element_1 *curr_element;
    struct list_element_1 *new_element;
    curr_element = hash_table[addr % HASHTABLE_SIZE].list_head;
    while (curr_element) {
        if (curr_element->address == addr) {
            curr_element->value = value;
            return;
        }
        curr_element = curr_element->next;
    }
    new_element = (struct list_element_1 *) malloc (sizeof (struct list_element_1));
    new_element->address = addr;
    new_element->value = value;
    new_element->next = hash_table[addr % HASHTABLE_SIZE].list_head;
    hash_table[addr % HASHTABLE_SIZE].list_head = new_element;
}

void remove_from_hash_table (type_hash_table *hash_table, bitsADDR addr) {
    struct list_element_1 *curr_element;
    struct list_element_1 *del_element;
    curr_element = hash_table[addr % HASHTABLE_SIZE].list_head;
    while (curr_element) {
        if (curr_element->address == addr) {
            del_element = curr_element;
            curr_element = curr_element->next;
            free (del_element);
            return;
        }
        curr_element = curr_element->next;
    } 
}


/*
 *
 * Procedures for working with tables of lists
 *
 */

void add_new_to_table_of_lists (
    type_table_of_lists *table_of_lists, 
    int i, 
    bitsADDR addr, 
    int total, 
    Thread_Time ENTER_time, 
    long int ENTER_pos)
{
    struct list_element_2 *new_element;
    new_element = (struct list_element_2 *) malloc (sizeof (struct list_element_2));
    new_element->address = addr;
    new_element->total = total;
    new_element->ENTER_time = ENTER_time;
    new_element->ENTER_pos = ENTER_pos;
    new_element->WAIT_time = 0;
    new_element->WAIT_pos = 0;
    if (table_of_lists[i].list_head == NULL) {
        new_element->next = NULL;
        table_of_lists[i].list_head = table_of_lists[i].list_tail = new_element;
    }
    else {
        new_element->next = table_of_lists[i].list_tail->next;
        table_of_lists[i].list_tail->next = new_element;
        table_of_lists[i].list_tail = new_element;
    }
}

void update_IN_data_last_table_of_lists (
    type_table_of_lists *table_of_lists, 
    int i, 
    int IN_order, 
    Thread_Time IN_time,
    Thread_Time original_IN_time)
{
    table_of_lists[i].list_tail->IN_order = IN_order;
    table_of_lists[i].list_tail->IN_time = IN_time;
    table_of_lists[i].list_tail->original_IN_time = original_IN_time;
}

void update_WAIT_data_last_table_of_lists (
    type_table_of_lists *table_of_lists, 
    int i, 
    Thread_Time WAIT_time,
    long int WAIT_pos)
{
    table_of_lists[i].list_tail->WAIT_time = WAIT_time;
    table_of_lists[i].list_tail->WAIT_pos = WAIT_pos;
}

void update_OUT_data_last_table_of_lists (
    type_table_of_lists *table_of_lists, 
    int i, 
    int OUT_order,
    Thread_Time OUT_time,
    Thread_Time original_OUT_time) 
{
    table_of_lists[i].list_tail->OUT_order = OUT_order;
    table_of_lists[i].list_tail->OUT_time = OUT_time;
    table_of_lists[i].list_tail->original_OUT_time = original_OUT_time;
}

void update_EXIT_data_last_table_of_lists (
    type_table_of_lists *table_of_lists, 
    int i, 
    Thread_Time EXIT_time,
    long int EXIT_pos) 
{
    table_of_lists[i].list_tail->EXIT_time = EXIT_time;
    table_of_lists[i].list_tail->EXIT_pos = EXIT_pos;
}

int read_IN_order_from_table_of_lists (type_table_of_lists *table_of_lists, int i, Thread_Time ENTER_time) {
    struct list_element_2 *curr_element;
    curr_element = table_of_lists[i].list_head;
    while (curr_element != table_of_lists[i].list_tail->next) {
        if (curr_element->ENTER_time == ENTER_time)
            return (curr_element->IN_order);
        curr_element = curr_element->next;
    } 
    return (0);
}

Thread_Time read_original_IN_time_from_table_of_lists (type_table_of_lists *table_of_lists, int i, Thread_Time ENTER_time) {
    struct list_element_2 *curr_element;
    curr_element = table_of_lists[i].list_head;
    while (curr_element != table_of_lists[i].list_tail->next) {
        if (curr_element->ENTER_time == ENTER_time)
            return (curr_element->original_IN_time);
        curr_element = curr_element->next;
    }
    return (0);
}

int read_OUT_order_from_table_of_lists (type_table_of_lists *table_of_lists, int i, Thread_Time ENTER_time) {
    struct list_element_2 *curr_element;
    curr_element = table_of_lists[i].list_head;
    while (curr_element != table_of_lists[i].list_tail->next) {
        if (curr_element->ENTER_time == ENTER_time)
            return (curr_element->OUT_order);
        curr_element = curr_element->next;
    } 
    return (0);
}

Thread_Time read_original_OUT_time_from_table_of_lists (type_table_of_lists *table_of_lists, int i, Thread_Time ENTER_time) {
    struct list_element_2 *curr_element;
    curr_element = table_of_lists[i].list_head;
    while (curr_element != table_of_lists[i].list_tail->next) {
        if (curr_element->ENTER_time == ENTER_time)
            return (curr_element->original_OUT_time);
        curr_element = curr_element->next;
    }
    return (0);
}

void read_element (
    type_table_of_lists *table_of_lists,
    int i, 
    Thread_Time ENTER_time,
    struct list_element_2 **element)
{
    struct list_element_2 *curr_element;

    *element = NULL;
    curr_element = table_of_lists[i].list_head;
    while (curr_element != table_of_lists[i].list_tail->next) {
        if (curr_element->ENTER_time == ENTER_time) {
            *element = curr_element;
            return;
        }
        curr_element = curr_element->next;
    }
}

void read_matching_IN_element_from_table_of_lists (
    type_table_of_lists *table_of_lists,
    int i, 
    bitsADDR addr, 
    int old_IN_order, 
    int new_IN_order, 
    Thread_Time original_IN_time,
    struct list_element_2 **match_element)
{
    struct list_element_2 *curr_element;

    *match_element = NULL;
    curr_element = table_of_lists[i].list_head;
    while (curr_element != table_of_lists[i].list_tail->next) {
        if ((curr_element->address == addr) && (curr_element->IN_order == new_IN_order))
            if (old_IN_order > new_IN_order)
                if (curr_element->original_IN_time <= original_IN_time)
                    *match_element = curr_element;
                else
                    return;
            else
                if (curr_element->original_IN_time >= original_IN_time) {
                    *match_element = curr_element;
                    return;
                }
        curr_element = curr_element->next;
    }
}

void read_matching_OUT_element_from_table_of_lists (
    type_table_of_lists *table_of_lists, 
    int i, 
    bitsADDR addr, 
    int old_OUT_order, 
    int new_OUT_order, 
    Thread_Time original_OUT_time,
    struct list_element_2 **match_element)
{
    struct list_element_2 *curr_element;
    *match_element = NULL;
    curr_element = table_of_lists[i].list_head;
    while (curr_element != table_of_lists[i].list_tail->next) {
        if ((curr_element->address == addr) && 
            (curr_element->OUT_order == new_OUT_order) &&
            (curr_element->WAIT_pos > 0))
            if (old_OUT_order > new_OUT_order)
                if (curr_element->original_OUT_time <= original_OUT_time) 
                    *match_element = curr_element;
                else
                    return;
            else
                if (curr_element->original_OUT_time >= original_OUT_time) {
                    *match_element = curr_element;
                    return;
                }
        curr_element = curr_element->next;
    } 
}


/* 
 *
 * Procedures for handling locks 
 *
 */

bitsLOCK read_lock (bitsADDR addr) {
    return(read_from_hash_table(locks_hash_table, addr));
}

void write_lock (bitsADDR addr, bitsLOCK value) {
    write_to_hash_table(locks_hash_table, addr, value);
}


/* 
 *
 * Procedures for handling barriers 
 *
 */

bitsLOCK read_barrier (bitsADDR addr) {
    return(read_from_hash_table(barriers_hash_table, addr));
}

void write_barrier (bitsADDR addr, bitsLOCK value) {
    write_to_hash_table(barriers_hash_table, addr, value);
}

void get_barrier (
    int i, 
    Thread_Time ENTER_time,
    struct list_element_2 **barrier)
{
    read_element (barrier_data, i, ENTER_time, barrier);
}

/* Next functions are crucial for finding matching threads in barriers */
void find_match_IN (
    int thread, 
    bitsADDR addr, 
    int new_IN_order, 
    Thread_Time ENTER_time,
    int threads_count,
    int *match_thread, 
    struct list_element_2 **match_barrier)
{
    int i,j;
    int matched_thread = 0;
    int old_IN_order;
    Thread_Time original_IN_time;
    struct list_element_2 *current_barrier;
    struct list_element_2 *matched_barrier=NULL;

    old_IN_order = read_IN_order_from_table_of_lists (barrier_data, thread, ENTER_time);
    original_IN_time = read_original_IN_time_from_table_of_lists (barrier_data, thread, ENTER_time);
    /* Find the first candidate on the barrier with that address */
    for (i=0; i<threads_count; i++) {
        read_matching_IN_element_from_table_of_lists (barrier_data, i, addr, old_IN_order, new_IN_order, original_IN_time, &current_barrier);
        if (current_barrier) {
            matched_thread = i;
            matched_barrier = current_barrier;
            break;
        }
    }
    /* Check if there is a closer barrier on that address among other threads */
    for (j=i+1; j<threads_count; j++) {
        read_matching_IN_element_from_table_of_lists (barrier_data, j, addr, old_IN_order, new_IN_order, original_IN_time, &current_barrier);
        if (current_barrier)
           if (abs(current_barrier->original_IN_time-original_IN_time) < abs(matched_barrier->original_IN_time-original_IN_time)) {
               matched_thread = j;
               matched_barrier = current_barrier;
           }
    }
    *match_thread = matched_thread;
    *match_barrier = matched_barrier;
}

void find_match_OUT (
    int thread, 
    bitsADDR addr, 
    int new_OUT_order, 
    Thread_Time ENTER_time,
    int threads_count,
    int *match_thread,
    struct list_element_2 **match_barrier)
{
    int i,j;
    int matched_thread = 0;
    int old_OUT_order;
    Thread_Time original_OUT_time;
    struct list_element_2 *current_barrier;
    struct list_element_2 *matched_barrier=NULL;

    old_OUT_order = read_OUT_order_from_table_of_lists (barrier_data, thread, ENTER_time);
    original_OUT_time = read_original_OUT_time_from_table_of_lists (barrier_data, thread, ENTER_time);
    /* Find the first candidate on the barrier with that address */
    for (i=0; i<threads_count; i++) {
        read_matching_OUT_element_from_table_of_lists (barrier_data, i, addr, old_OUT_order, new_OUT_order, original_OUT_time, &current_barrier);
        if (current_barrier) {
            matched_thread = i;
            matched_barrier = current_barrier;
            break;
        }
    }
    /* Check if there is a closer barrier on that address among other threads */
    for (j=i+1; j<threads_count; j++) {
        read_matching_OUT_element_from_table_of_lists (barrier_data, j, addr, old_OUT_order, new_OUT_order, original_OUT_time, &current_barrier);
        if (current_barrier)
           if (abs(current_barrier->original_OUT_time-original_OUT_time) < abs(matched_barrier->original_OUT_time-original_OUT_time)) {
               matched_thread = j;
               matched_barrier = current_barrier;
           }
    }
    *match_thread = matched_thread;
    *match_barrier = matched_barrier;
}
